from typing import Any, Callable, Dict

import pandas as pd
import pyarrow
from pyarrow import parquet

from common_client_scheduler import (
    Display,
    ExportResponse,
    StructRef,
    PandasIndexMetadata,
    PandasSeriesMetadata,
    PandasDFMetadata,
    NDArrayMetadata,
)
from common_client_scheduler.requests_responses import AwsCredentials
from common_client_scheduler.structs import IndexType, ParquetMetadata
from terality_serde import StructType
from terality_serde.json_encoding import (
    _endecode_json_indices_inplace,
    _endecode_json_columns_inplace,
)

from .. import DataTransfer, copy_to_user_s3_bucket
from ..utils.azure import test_for_azure_libs, parse_azure_filesystem
from .. import global_client
from ...exceptions import TeralityClientError


def _deserialize_display(
    aws_credentials: AwsCredentials, to_display: Display
):  # pylint: disable=unused-argument,useless-return
    # No need to force it in package dependencies, if it gets called it means we are in a Jupyter Notebook
    # and and this dependency is present
    # noinspection PyUnresolvedReferences
    from IPython.display import display, HTML

    display(HTML(to_display.value))


def _deserialize_export(aws_credentials: AwsCredentials, export: ExportResponse) -> None:
    path = export.path
    transfer_id = export.transfer_id
    if path.startswith("s3://"):
        copy_to_user_s3_bucket(aws_credentials, transfer_id, export.aws_region, path)
    elif path.startswith("abfs://") or path.startswith("az://"):
        test_for_azure_libs()
        from ..data_transfers.azure import copy_to_azure_datalake

        storage_account_name, container, path = parse_azure_filesystem(path, export.storage_options)
        copy_to_azure_datalake(
            aws_credentials=aws_credentials,
            transfer_id=transfer_id,
            aws_region=export.aws_region,
            storage_account_name=storage_account_name,
            container=container,
            path_template=path,
        )
    elif path.startswith("gs://"):
        raise TeralityClientError(
            "Currently, Terality does not support GCP storage. Please use AWS S3 or Azure paths."
        )
    else:
        config = global_client().get_download_config()
        DataTransfer.download_to_local_files(
            config,
            aws_credentials,
            transfer_id,
            path,
            export.is_folder,
            export.with_leading_zeros,
        )


def _download_df(
    aws_credentials: AwsCredentials,
    transfer_id: str,
    metadata: ParquetMetadata,
) -> pd.DataFrame:
    client = global_client().get_download_config()
    bytes_io = DataTransfer.download_to_bytes(client, aws_credentials, transfer_id)

    df = pd.read_parquet(bytes_io)
    # range index get dropped by read_parquet on a df with no cols
    if df.shape[0] == 0:
        file_schema = parquet.read_schema(pyarrow.BufferReader(bytes_io.getvalue()))
        pandas_metadata = file_schema.pandas_metadata
        index_metadata = pandas_metadata["index_columns"]
        if len(index_metadata) == 1 and isinstance(index_metadata[0], dict):
            range_index = index_metadata[0]
            if range_index["kind"] == "range":
                index = pd.RangeIndex(
                    start=range_index["start"],
                    stop=range_index["stop"],
                    step=range_index["step"],
                    name=range_index["name"],
                )
                return pd.DataFrame(index=index)
    _endecode_json_indices_inplace(df, metadata.index.encodings, encode=False)
    _endecode_json_columns_inplace(df, metadata.cols.encodings, encode=False)

    df.columns = metadata.cols.col_names
    # mypy does not understand that it's not a basic setattre and that df.columns is not a list of str
    df.columns.name = metadata.cols.col_names_name  # type: ignore

    if isinstance(df.index, pd.MultiIndex):
        df.index.names = metadata.index.names
    df.index.name = metadata.index.name
    if metadata.index.datetime_index_freq is not None:
        assert isinstance(df.index, pd.DatetimeIndex)
        df.index.freq = metadata.index.datetime_index_freq

    return df


def _download_ndarray_from_metadata(
    aws_credentials: AwsCredentials, ndarray_metadata: NDArrayMetadata
):
    # noinspection PyTypeChecker
    # TODO use np.load, but we have to upload accordingly within scheduler ndarray_to_numpy_metadata.
    df = _download_df(
        aws_credentials, ndarray_metadata.transfer_id, ndarray_metadata.parquet_metadata
    )
    assert len(df.columns) == 1
    return df.iloc[:, 0].to_numpy()


def _download_index_from_metadata(
    aws_credentials: AwsCredentials, index_metadata: PandasIndexMetadata
) -> pd.Index:
    df = _download_df(aws_credentials, index_metadata.transfer_id, index_metadata.parquet_metadata)
    return df.index


def _download_series_from_metadata(
    aws_credentials: AwsCredentials, series_metadata: PandasSeriesMetadata
) -> pd.Series:
    df = _download_df(
        aws_credentials, series_metadata.transfer_id, series_metadata.parquet_metadata
    )
    series = df.iloc[:, 0]
    return series


def _download_df_from_metadata(
    aws_credentials: AwsCredentials, df_metadata: PandasDFMetadata
) -> pd.DataFrame:
    df = _download_df(aws_credentials, df_metadata.transfer_id, df_metadata.parquet_metadata)
    return df


_decoder: Dict[Any, Callable[[AwsCredentials, Any], Any]] = {
    NDArrayMetadata: _download_ndarray_from_metadata,
    PandasIndexMetadata: _download_index_from_metadata,
    PandasSeriesMetadata: _download_series_from_metadata,
    PandasDFMetadata: _download_df_from_metadata,
    Display: _deserialize_display,
    ExportResponse: _deserialize_export,
}


def decode(credentials_fetcher: AwsCredentials, obj):  # pylint: disable=invalid-name
    from terality import (
        NDArray,
        Index,
        Int64Index,
        Float64Index,
        DatetimeIndex,
        MultiIndex,
        Series,
        DataFrame,
    )  # To avoid circular dependencies
    from terality._terality.terality_structures import DataFrameGroupBy, SeriesGroupBy

    structs = {
        IndexType.INDEX: Index,
        IndexType.INT64_INDEX: Int64Index,
        IndexType.FLOAT64_INDEX: Float64Index,
        IndexType.DATETIME_INDEX: DatetimeIndex,
        IndexType.MULTI_INDEX: MultiIndex,
        StructType.NDARRAY: NDArray,
        StructType.SERIES: Series,
        StructType.DATAFRAME: DataFrame,
        StructType.SERIES_GROUPBY: SeriesGroupBy,
        StructType.DATAFRAME_GROUPBY: DataFrameGroupBy,
    }

    if isinstance(obj, StructRef):
        # noinspection PyProtectedMember
        # NOTE: hash(StrEnum.X) = hash(StrEnum.X.value)
        return structs[obj.type]._new(id_=obj.id)  # type: ignore
    if type(obj) in _decoder:
        return _decoder[type(obj)](credentials_fetcher, obj)
    return obj



from typing import Optional


class TeralitySerdeError(Exception):
    pass


class TeralitySerializationError(TeralitySerdeError):
    def __init__(self, message: str, unserializable_type: Optional[str] = None) -> None:
        super().__init__(message)
        self.unserializable_type = unserializable_type


class TeralityDeserializationError(TeralitySerdeError):
    pass



# All imports start with an underscore, except the one we
# want to publicly export.
import inspect as _inspect
import importlib as _importlib
import sys as _sys

import pandas as _pd

from terality_serde.json_encoding import (
    instantiate_pandas_arrow_extension_types as _instantiate_pandas_arrow_extension_types,
)

from .version import __version__

# Rename all imported symbols that should not be public to start with an underscore.
# (prevent a user from doing `from terality import top_level_functions` for instance).
from ._terality.terality_structures.top_level import (
    get_dynamic_top_level_attribute as _get_dynamic_top_level_attribute,
    top_level_functions as _top_level_functions,
)
from ._terality.utils.anonymous import create_anonymous_user as _create_anonymous_user

from ._terality.terality_structures.ndarray import NDArray
from ._terality.terality_structures.dataframe import DataFrame
from ._terality.terality_structures.index import Index
from ._terality.terality_structures.index import Int64Index
from ._terality.terality_structures.index import Float64Index
from ._terality.terality_structures.index import DatetimeIndex
from ._terality.terality_structures.index import MultiIndex
from ._terality.terality_structures.series import Series


def quickstart():
    """Get started with Terality without any setup: create a temporary Terality account on this machine.

    By using this method, you agree to the terms and conditions at https://www.terality.com/terms.

    This method should only be used when evaluating Terality. For any real usage, please create a permanent
    Terality account at https://app.terality.com (it's free).

    The account created by this quickstart method has a limited duration, after which all data in this
    account will be deleted. This quickstart account also has a lower data cap than a normal account, and
    can't be upgraded.

    When the lifetime of the quickstart account is exceeded, further calls to Terality will return errors.
    """
    _create_anonymous_user()


def __getattr__(attribute: str):
    return _get_dynamic_top_level_attribute(attribute)


def __dir__():
    # Static members.
    members = set(name for name in globals() if not name.startswith("_"))
    # Dynamic members.
    members |= _top_level_functions
    return members


# Reexport all (public) pandas submodules too.
# This is required to make importing submodules ("import terality.tseries.offsets") work.
for _attribute_name in dir(_pd):
    # Don't reexport private pandas names
    if _attribute_name.startswith("_"):
        continue
    # pandas does some magic when importing these submodules, for instance:
    #     >>> import pandas
    #     >>> from pandas import pandas as pandas_2
    #     >>> pandas_2 is pandas
    #     True
    # so let's just skip them instead of trying to work around this
    if _attribute_name in ["offsets", "pandas"]:
        continue
    _pd_attribute = getattr(_pd, _attribute_name)
    if _inspect.ismodule(_pd_attribute):
        _sys.modules[f"terality.{_attribute_name}"] = _importlib.import_module(
            f".{_attribute_name}", "pandas"
        )


_instantiate_pandas_arrow_extension_types()



"""Main entry point for the Terality command line interface."""
from terality._terality.client import anonymous_client
import click

from common_client_scheduler.requests_responses import CreateUserRequest, CreateUserResponse
from terality._terality.utils.account import (
    has_valid_terality_config_file,
    has_valid_terality_credentials_file,
)
from terality._terality.utils.config import TeralityConfig, TeralityCredentials


@click.group(
    help="This is the Terality command line interface. Documentation is available at https://docs.terality.com.",
    epilog="Terality is currently beta-quality software. For any support question, please write to support@terality.com.",
)
def cli():
    pass


@cli.group(name="account", short_help="Account management utilities.")
def account():
    pass


@account.command(
    name="create",
    short_help="Register a new Terality account.",
    help="Create a free Terality account. We recommend using https://app.terality.com/ instead, as it will provide better account management features. This command may be removed in future Terality versions.",
)
@click.option(
    "--email", prompt="Your email address", confirmation_prompt="Please confirm your email address"
)
@click.option(
    "--full-name", prompt="Your full name (optional)", default="", help="Your full name (optional)."
)
@click.option(
    "--company",
    prompt="Your company name (optional)",
    default="",
    help="Your company name (optional).",
)
@click.option(
    "--skip-configuration", help="Do not create a Terality configuration file.", is_flag=True
)
def account_create(email: str, full_name: str, company: str, skip_configuration: bool):
    click.echo("")
    click.echo(
        "To continue, you must read and accept the Terality Terms of Service, available at https://www.terality.com/terms, and the Terality Privacy Policy, available at https://www.terality.com/privacy."
    )
    click.echo("Enter 'y' to indicate that you accept these terms.")
    click.confirm("Do you accept the Terality Terms of Service?", abort=True, default=True)
    click.confirm("Do you accept the Terality Privacy Policy?", abort=True, default=True)

    click.echo("")
    click.echo("Your account information:")
    click.echo("")
    click.echo(f"Email: {email}")
    if full_name:
        click.echo(f"Full name: {full_name}")
    if company:
        click.echo(f"Company name: {company}")
    click.echo("")

    click.confirm("Confirm and create account?", abort=True, default=True)

    payload = CreateUserRequest(
        email=email,
        user_accepted_privacy_policy=True,
        user_accepted_terms_of_service=True,
        company=company,
        full_name=full_name,
    )
    client = anonymous_client()
    res = client.poll_for_answer("users/", payload=payload)
    if not isinstance(
        res, CreateUserResponse
    ):  # TODO: this validation should be handled at the deserialization level
        raise ValueError(
            f"Unexpected server answer: expected type 'CreateUserResponse', got '{type(res)}'"
        )

    click.echo("")
    click.echo("Congratulations! Your Terality account has been created.")
    click.echo("Your API key is: ")
    click.echo("")
    click.echo("    " + res.api_key)
    click.echo("")
    click.echo(
        "Keep it in a safe place (such as your password manager). If you lose it, please reach out to support@terality.com to renew it."
    )
    click.echo("")

    if not skip_configuration:
        _write_terality_configuration(email, res.api_key)
        click.echo(
            f"Your Terality credentials have been automatically written to {TeralityCredentials.file_path()}. "
            "You can now start using Terality!"
        )

    click.echo("For documentation and resources, go to https://docs.terality.com.")


# TODO: add default from configuration files (show_default and default options in click)
# but this requires cleaning up the config classes beforehand.
@account.command(
    name="configure",
    short_help="Configure this system to use your Terality account",
    help="Create the Terality configuration files on this system. It's recommended not to provide the --api-key flag on the command line, and enter it interactively instead, to avoid leaking it in your shell history.",
)
@click.option(
    "--email",
    prompt="Your email address",
    help="Email address associated with the Terality account.",
)
@click.option("--api-key", prompt="Your Terality API key", help="Terality API key.")
@click.option(
    "--overwrite", help="Overwrite existing configuration files without confirmation.", is_flag=True
)
def account_configure(email: str, api_key: str, overwrite: bool):
    has_config = has_valid_terality_config_file()
    has_credentials = has_valid_terality_credentials_file()

    if has_config:
        _print_warning(
            f"A Terality configuration file is already present at {TeralityConfig.file_path()}."
        )
    if has_credentials:
        _print_warning(
            f"A Terality credentials file is already present at {TeralityCredentials.file_path()}."
        )

    if (has_config or has_credentials) and not overwrite:
        click.confirm(
            "Are you sure you want to overwrite the existing Terality configuration?", abort=True
        )

    config = TeralityConfig()
    credentials = TeralityCredentials(user_id=email, user_password=api_key)
    config.save()
    credentials.save()

    click.echo("Terality account succesfully configured on this system.")


@account.command(
    name="info",
    short_help="Display the current Terality configuration.",
    help="Display the Terality configuration on this system.",
)
def account_info():
    try:
        config = TeralityConfig().load()
        config_is_available = True
    except Exception:
        # TODO: fine grained handling of exceptions (display accurate error)
        config_is_available = False
        config = None
    try:
        credentials = TeralityCredentials.load()
        credentials_are_available = True
    except Exception:
        # TODO: fine grained handling of exceptions (display accurate error)
        credentials_are_available = False
        credentials = None

    if not config_is_available:
        _print_warning(
            f"Could not read the Terality configuration file at {TeralityConfig.file_path()}."
        )
    if not credentials_are_available:
        _print_warning(
            f"Could not read the Terality credentials file at {TeralityCredentials.file_path()}."
        )

    if not config_is_available or not credentials_are_available:
        click.echo("You can configure your Terality account by running:")
        click.echo("")
        click.echo("    terality account configure")
        click.echo("")
        return

    assert config is not None
    assert credentials is not None
    click.echo("Account")
    click.echo(f"    User email: {credentials.user_id}")
    click.echo("    User API key: <hidden>")
    click.echo("")
    click.echo("Advanced")
    click.echo(
        f"    Configuration files: {TeralityConfig.file_path()}, {TeralityCredentials.file_path()}"
    )
    click.echo(f"    Terality instance (API URL): {config.full_url()}")


def _print_warning(message: str):
    click.echo(click.style("WARNING:", fg="yellow") + " " + message)


def _write_terality_configuration(email: str, api_key: str):
    credentials = TeralityCredentials(user_id=email, user_password=api_key)
    credentials.save()
    config = TeralityConfig()
    config.save()


if __name__ == "__main__":
    cli()



import pandas as pd
from terality_serde import StructType

from . import ClassMethod, Struct
from .structure import StructIterator


class ClassMethodSeries(ClassMethod):
    _class_name: StructType = StructType.SERIES
    _pandas_class = pd.Series
    _additional_class_methods = ClassMethod._additional_class_methods | {
        "random",
        "random_integers",
    }


class Series(Struct, metaclass=ClassMethodSeries):
    _pandas_class_instance = pd.Series(dtype="float64")
    _accessors = {"str", "dt"}
    _additional_methods = Struct._additional_methods | {"get_range_auto", "random"}

    def __iter__(self):
        return StructIterator(self)

    def to_dict(self, into: type = dict):
        pd_series = self._call_method(None, "to_pandas")
        return pd_series.to_dict(into=into)

    def to_list(self):
        pd_series = self._call_method(None, "to_pandas")
        return pd_series.to_list()

    def tolist(self):
        return self.to_list()



from dataclasses import dataclass
from typing import Hashable, List, Optional

from terality_serde import SerdeMixin, IndexType
from terality_serde.json_encoding import ArrayEncoding


@dataclass
class StructRef(SerdeMixin):
    type: str  # strings from IndexType/StructType enums
    id: str  # pylint: disable=invalid-name


@dataclass
class ParquetMetadataIndex(SerdeMixin):
    type_: IndexType
    encodings: List[Optional[ArrayEncoding]]
    names: List[Hashable]
    name: Hashable
    datetime_index_freq: Optional[str]


@dataclass
class ParquetMetadataCols(SerdeMixin):
    encodings: List[Optional[ArrayEncoding]]
    col_names: List[Hashable]
    col_names_name: Hashable


@dataclass
class ParquetMetadata(SerdeMixin):
    """
    When to_pandas/from_pandas methods are performed, these metadata
    are sent between client/scheduler instead of raw data.
    """

    index: ParquetMetadataIndex
    cols: ParquetMetadataCols


@dataclass
class NDArrayMetadata(SerdeMixin):
    transfer_id: str
    parquet_metadata: ParquetMetadata


@dataclass
class PandasIndexMetadata(SerdeMixin):
    transfer_id: str
    parquet_metadata: ParquetMetadata


@dataclass
class PandasSeriesMetadata(SerdeMixin):
    transfer_id: str
    parquet_metadata: ParquetMetadata


@dataclass
class PandasDFMetadata(SerdeMixin):
    transfer_id: str
    parquet_metadata: ParquetMetadata


@dataclass
class Display(SerdeMixin):
    value: str


@dataclass
class PandasFunctionRequest(SerdeMixin):
    function_type: str
    function_accessor: Optional[str]
    function_name: str
    args: str
    kwargs: str



from dataclasses import dataclass


@dataclass
class TeralityError(Exception):
    """Base class for errors in the client"""

    message: str

    def __str__(self):
        return self.message


class TeralityClientError(TeralityError):
    """A client error occured. Retrying won't help."""



"""Endpoints requests/responses."""
from __future__ import annotations
from dataclasses import dataclass
import datetime as dt
from typing import Any, Dict, List, Optional, Union

from terality_serde import SerdeMixin, SerializableEnum

from . import TransferConfig, TransferConfigLocal


@dataclass
class PendingComputationResponse(SerdeMixin):
    pending_computation_id: str


@dataclass
class ComputationResponse(SerdeMixin):
    result: Any
    inplace: bool


@dataclass
class CreateSessionResponse(SerdeMixin):
    id: str  # pylint: disable=invalid-name
    upload_config: Union[TransferConfig, TransferConfigLocal]
    download_config: Union[TransferConfig, TransferConfigLocal]


@dataclass
class DeleteSessionResponse(SerdeMixin):
    """Response to the delete session endpoint"""


@dataclass
class Upload(SerdeMixin):
    path: str


@dataclass
class UploadRequest(SerdeMixin):
    path: str
    transfer_id: str
    aws_region: Optional[str]

    def terality_hash(self):
        return self.path


@dataclass
class ExportRequest(SerdeMixin):
    # Parameters of the "to_xxx" (to_csv, to_parquet...) function.
    # The server will echo it back in the response, but (as of now) don't use this information otherwise.
    path: str
    storage_options: Optional[Dict[str, Any]]
    # Preferred region for the Terality export bucket.
    aws_region: Optional[str]


@dataclass
class ExportResponse(SerdeMixin):
    path: str
    storage_options: Optional[Dict[str, Any]]
    aws_region: Optional[str]
    transfer_id: str

    is_folder: bool  # True means it is the result of to_csv_folder or to_parquet_folder
    with_leading_zeros: bool = False  # Only relevant if is_folder is True. Output filename numbers will have leading zeros.


@dataclass
class AwsCredentials(SerdeMixin):
    access_key_id: str
    secret_access_key: str
    session_token: str


@dataclass
class DataTransferResponse(SerdeMixin):
    temporary_upload_aws_credentials: AwsCredentials


class StorageService(SerializableEnum):
    AWS_S3 = "AWS_S3"
    AZURE_BLOB_STORAGE = "AZURE_BLOB_STORAGE"


@dataclass
class AwsPresignedUrlSource(SerdeMixin):
    """A presigned AWS URL allowing to run GetObject on an S3 object part."""

    # Possible optimization: for the same object (bucket + key), the same presigned URL could
    # be reused for each range to get (the same presigned URL can be requested with different
    # Range HTTP headers).
    url: str
    object_key: ObjectStorageKey
    # warning, these offsets are inclusive
    range_start_byte: int
    range_end_byte: int


@dataclass(frozen=True, order=True)
class ObjectStorageKey(SerdeMixin):
    """Generic representation of the path to an object (= blob in Azure) in some cloud storage service.

    (AWS Simple Storage Service (S3), Google Cloud Storage (GCS), Azure Blob Storage...)
    """

    bucket: str  # maps to a container in Azure Blob Storage, called "bucket" in most of the other services
    # (Azure Blob Storage also needs an extra "storage account name" information to uniquely identity
    # the blob. It should be transmitted along this object).

    key: str  # also called an "object name" in Google Cloud Storage and a "blob name" in Azure


@dataclass
class ImportFromS3Source(SerdeMixin):
    presigned_urls: List[AwsPresignedUrlSource]


@dataclass
class AzureBlobSource(SerdeMixin):
    """A reference to an blob stored in Azure Blob storage, plus a shared access signature (SAS) that can read it.

    The SAS is a temporary credential generated by the Terality client that can be used as a credential
    to get read-only access to the given blob.
    """

    object_key: ObjectStorageKey
    shared_access_signature: str
    storage_account_name: str
    # Possible optimization: for the same object (container + name), the same SAS could
    # be reused for each range to get.
    range_start_byte: int
    range_end_byte: int


# Even if the user provided an Azure Datalake Gen2 URL, the server always work
# at the Blob Storage level for import/exports. The client does the mapping.
@dataclass
class ImportFromAzureBlobStorageSource(SerdeMixin):
    objects: List[AzureBlobSource]


@dataclass
class ImportFromCloudRequest(SerdeMixin):
    service: StorageService
    source: Union[ImportFromS3Source, ImportFromAzureBlobStorageSource]


@dataclass
class ImportFromCloudResponse(SerdeMixin):
    transfer_id: str


@dataclass
class AwsS3ObjectPartExportRequest(SerdeMixin):
    """A presigned AWS URL allowing to run UploadPart on a S3 object part."""

    # Source
    source_object_key: ObjectStorageKey
    range_start_byte: int
    range_end_byte: int

    # Destination
    presigned_url: str

    # The server doesn't need to know the following pieces of information, but it makes writing the client
    # easier. Instead of matching the server results back using the URL (or an arbitrary ID), the server can
    # send back these IDs and the client can directly use them to complete the upload.
    # Note that the presigned URL already contains this in cleartext (we don't want want to rely on it,
    # but it shows that we don't have to consider this information as sensitive).
    multipart_upload_id: str
    part_number: int
    destination_object_key: ObjectStorageKey


@dataclass
class AzureBlobStoragePartExportRequest(SerdeMixin):
    # Source
    source_object_key: ObjectStorageKey  # S3 object in the Terality export bucket
    range_start_byte: int
    range_end_byte: int

    # Destination
    destination_object_key: ObjectStorageKey  # (container, name) of an Azure Blob in storage_account_name
    storage_account_name: str
    shared_access_signature: str


@dataclass
class AwsS3PartsExport(SerdeMixin):
    parts: List[AwsS3ObjectPartExportRequest]


@dataclass
class AzureBlobStoragePartsExport(SerdeMixin):
    parts: List[AzureBlobStoragePartExportRequest]


@dataclass
class ExportToCloudRequest(SerdeMixin):
    service: StorageService
    export_request: Union[AwsS3PartsExport, AzureBlobStoragePartsExport]


@dataclass
class UploadedS3PartInfo(SerdeMixin):
    destination_object_key: ObjectStorageKey
    etag: str
    part_number: int
    multipart_upload_id: str


@dataclass
class ExportToS3Response(SerdeMixin):
    uploaded_parts: List[UploadedS3PartInfo]


@dataclass
class ExportToAzureBlobStorageResponse(SerdeMixin):
    pass


@dataclass
class CreateUserRequest(SerdeMixin):
    email: str
    user_accepted_privacy_policy: bool
    user_accepted_terms_of_service: bool
    full_name: Optional[str]
    company: Optional[str]


@dataclass
class CreateUserResponse(SerdeMixin):
    api_key: str


@dataclass
class ErrorResponse(SerdeMixin):
    message: str


@dataclass
class ClientErrorContext(SerdeMixin):
    code: str
    filename: str
    exception_str: Optional[str]


@dataclass
class CreateAnonymousUserResponse(SerdeMixin):
    user_id: str
    api_key: str
    expires_at: dt.datetime



from typing import Set

import numpy as np
from terality_serde import StructType

from . import ClassMethod, Struct, StructIterator


class ClassMethodNDArray(ClassMethod):
    _class_name: str = StructType.NDARRAY
    _pandas_class = np.ndarray
    _additional_class_methods: Set[str] = {"from_numpy"}


class NDArray(Struct, metaclass=ClassMethodNDArray):
    """
    np.ndarray distributed version. This class aims to
    be able to manipulate ndarray during pandas workflow,
    but numpy API is not supported.
    Thus, any method call on an NDArray instance will fail,
    and the constructor is not provided.
    """

    _pandas_class_instance = np.empty(0)
    _additional_methods: Set[str] = {"to_numpy"}
    _indexers: Set[str] = set()

    def __iter__(self):
        return StructIterator(self)



from io import TextIOBase
from pathlib import Path
from typing import Optional, Union


def write_output(
    content: str,
    output: Optional[Union[TextIOBase, str, Path]] = None,
    encoding: Optional[str] = None,
) -> None:
    if isinstance(output, (str, Path)):
        with open(output, "w", encoding=encoding) as file:
            file.write(content)
        return

    if isinstance(output, TextIOBase):
        output.write(content)
        return

    raise ValueError(f"Unsupported output type: {type(output)}")



from typing import Dict, Optional, List, Sequence, Hashable, Union

import pandas as pd

from common_client_scheduler import ExportRequest
from terality.exceptions import TeralityClientError
from terality_serde import StructType

from . import ClassMethod, Struct
from ..data_transfers.common import S3


class ClassMethodDF(ClassMethod):
    _class_name: str = StructType.DATAFRAME
    _pandas_class = pd.DataFrame
    _additional_class_methods = ClassMethod._additional_class_methods | {
        "from_dict",
        "from_records",
    }


def _make_export_request(path: str, storage_options: Optional[Dict] = None) -> ExportRequest:
    if not isinstance(path, str):
        raise TeralityClientError("Export methods only support a str path.")

    if path.startswith("s3://"):
        bucket = path[5:].split("/", maxsplit=1)[0]
        aws_region = S3.client().get_bucket_location(Bucket=bucket)["LocationConstraint"]
    else:
        aws_region = None
    return ExportRequest(path=path, aws_region=aws_region, storage_options=storage_options)


class DataFrame(Struct, metaclass=ClassMethodDF):
    _class_name: str = StructType.DATAFRAME
    _pandas_class_instance = pd.DataFrame()
    _additional_methods = Struct._additional_methods | {
        "to_csv_folder",
        "to_parquet_folder",
    }
    _args_to_replace: dict = {
        # "to_csv" : (0, ExportRequest),
        "to_excel": (0, _make_export_request)
    }

    def _on_missing_attribute(self, item: str):
        return self._call_method(None, "df_col_by_attribute_access", item)

    def __iter__(self):
        # Iterating on a `DataFrame` is the same as iterating on its columns.
        return self.columns.__iter__()

    def to_csv(  # pylint: disable=too-many-arguments, too-many-locals
        self,
        path_or_buf: str,
        sep: str = ",",
        na_rep: str = "",
        float_format: Optional[str] = None,
        columns: Optional[Sequence[Optional[Hashable]]] = None,
        header: Union[bool, List[str]] = True,
        index: bool = True,
        index_label: Optional[Union[str, Sequence, bool]] = None,
        mode: str = "w",
        encoding: Optional[str] = None,
        compression: Union[str, dict] = "infer",
        quoting: Optional[int] = None,
        quotechar: str = '"',
        line_terminator: Optional[str] = None,
        chunksize: Optional[int] = None,
        date_format: Optional[str] = None,
        doublequote: bool = True,
        escapechar: Optional[str] = None,
        decimal: str = ".",
        errors: str = "strict",
        storage_options: Optional[Dict] = None,
    ) -> Optional[str]:
        export_request = _make_export_request(path_or_buf, storage_options)
        return self._call_method(
            None,
            "to_csv",
            export_request,
            sep=sep,
            na_rep=na_rep,
            float_format=float_format,
            columns=columns,
            header=header,
            index=index,
            index_label=index_label,
            mode=mode,
            encoding=encoding,
            compression=compression,
            quoting=quoting,
            quotechar=quotechar,
            line_terminator=line_terminator,
            chunksize=chunksize,
            date_format=date_format,
            doublequote=doublequote,
            escapechar=escapechar,
            decimal=decimal,
            errors=errors,
            storage_options=storage_options,
        )

    def to_csv_folder(  # pylint: disable=too-many-arguments, too-many-locals
        self,
        path_or_buf: str,
        num_files: Optional[int] = None,
        num_rows_per_file: Optional[int] = None,
        in_memory_file_size: Optional[int] = None,
        with_leading_zeros: bool = False,
        sep: str = ",",
        na_rep: str = "",
        float_format: Optional[str] = None,
        columns: Optional[Sequence[Optional[Hashable]]] = None,
        header: Union[bool, List[str]] = True,
        index: bool = True,
        index_label: Optional[Union[str, Sequence, bool]] = None,
        mode: str = "w",
        encoding: Optional[str] = None,
        compression: Union[str, dict] = "infer",
        quoting: Optional[int] = None,
        quotechar: str = '"',
        line_terminator: Optional[str] = None,
        chunksize: Optional[int] = None,
        date_format: Optional[str] = None,
        doublequote: bool = True,
        escapechar: Optional[str] = None,
        decimal: str = ".",
        errors: str = "strict",
        storage_options: Optional[Dict] = None,
    ) -> Optional[str]:
        """
        Store the DataFrame in several CSV files. Exactly one of num_files, num_rows_per_file, file_size must be provided.
        The number of files to produce is deduced from the parameter filled.
        The path basename must contain the character *, that will be replaced by file number when producing files.
        If the path contains a non-existing folder, it is created.
        Leading zeros can be added to file numbers so each filename will have the same length.

        NOTE: This method has a specific documentation https://docs.terality.com/getting-terality/api-reference/write-to-multiple-files
        which should be updated if needed.
        """

        export_request = _make_export_request(path_or_buf, storage_options)
        return self._call_method(
            None,
            "to_csv_folder",
            export_request,
            num_files=num_files,
            num_rows_per_file=num_rows_per_file,
            in_memory_file_size=in_memory_file_size,
            with_leading_zeros=with_leading_zeros,
            sep=sep,
            na_rep=na_rep,
            float_format=float_format,
            columns=columns,
            header=header,
            index=index,
            index_label=index_label,
            mode=mode,
            encoding=encoding,
            compression=compression,
            quoting=quoting,
            quotechar=quotechar,
            line_terminator=line_terminator,
            chunksize=chunksize,
            date_format=date_format,
            doublequote=doublequote,
            escapechar=escapechar,
            decimal=decimal,
            errors=errors,
            storage_options=storage_options,
        )

    def to_parquet(
        self,
        path: str,
        engine: str = "auto",
        compression: Optional[str] = "snappy",
        index: Optional[bool] = None,
        partition_cols: Optional[List[str]] = None,
        storage_options: Optional[Dict] = None,
    ):

        return self._call_method(
            None,
            "to_parquet",
            _make_export_request(path, storage_options),
            engine=engine,
            compression=compression,
            index=index,
            partition_cols=partition_cols,
            storage_options=storage_options,
        )

    def to_parquet_folder(  # pylint: disable=too-many-arguments
        self,
        path: str,
        num_files: Optional[int] = None,
        num_rows_per_file: Optional[int] = None,
        in_memory_file_size: Optional[int] = None,
        with_leading_zeros: bool = False,
        engine: str = "auto",
        compression: Optional[str] = "snappy",
        index: Optional[bool] = None,
        partition_cols: Optional[List[str]] = None,
        storage_options: Optional[Dict] = None,
    ):
        """
        Store the DataFrame in several parquet files. Exactly one of num_files, num_rows_per_file, file_size must be provided.
        The number of files to produce is deduced from the parameter filled.
        The path basename must contain the character *, that will be replaced by file number when producing files.
        If the path contains a non-existing folder, it is created.
        Leading zeros can be added to file numbers so each filename will have the same length.

        NOTE: This method has a specific documentation https://docs.terality.com/getting-terality/api-reference/write-to-multiple-files
        which should be updated if needed.
        """

        return self._call_method(
            None,
            "to_parquet_folder",
            _make_export_request(path, storage_options),
            engine=engine,
            compression=compression,
            index=index,
            partition_cols=partition_cols,
            storage_options=storage_options,
            num_files=num_files,
            num_rows_per_file=num_rows_per_file,
            in_memory_file_size=in_memory_file_size,
            with_leading_zeros=with_leading_zeros,
        )

    def to_dict(self, orient: str = "dict", into: type = dict):
        pd_df = self._call_method(None, "to_pandas")
        return pd_df.to_dict(orient=orient, into=into)

    def info(
        self,
        verbose: Optional[bool] = None,
        max_cols: Optional[int] = None,
        memory_usage: Optional[Union[bool, str]] = None,
        null_counts: Optional[bool] = None,
    ):

        info = self._call_method(
            None,
            "info",
            verbose=verbose,
            max_cols=max_cols,
            memory_usage=memory_usage,
            null_counts=null_counts,
        )
        print(info)



from __future__ import annotations
from abc import ABC, abstractmethod
from typing import ByteString, List, Type


class ProtobufWrapper(ABC):
    """A Wrapper around a protobuf object for efficient serialization"""

    _protobuf_class: Type
    subclasses: List[Type[ProtobufWrapper]] = []

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        cls.subclasses.append(cls)

    @property
    @abstractmethod
    def proto(self):
        """Returns protobuf object wrapping containing current object's data"""

    @classmethod
    @abstractmethod
    def from_proto(cls, proto) -> ProtobufWrapper:
        """Returns an instance of the ProtobufWrapper subclass initialized using the data provided inside the proto object"""

    def serialize(self) -> ByteString:
        return self.proto.SerializeToString()

    @classmethod
    def parse(cls, bytes_: ByteString) -> ProtobufWrapper:
        proto = cls._protobuf_class()
        proto.ParseFromString(bytes_)
        return cls.from_proto(proto)



from typing import Any, Awaitable, Callable


def apply_func_on_object_recursively(obj: Any, f: Callable, apply_key_dict: bool = False) -> Any:
    """Recursively apply a callable to a JSON-like tree of objects (i.e. a combination of dictionaries and lists)"""
    if isinstance(obj, list):
        return [apply_func_on_object_recursively(value, f, apply_key_dict) for value in obj]
    if isinstance(obj, tuple):  # pylint: disable=no-else-return
        return tuple(  # pylint: disable=consider-using-generator
            [apply_func_on_object_recursively(value, f, apply_key_dict) for value in obj]
        )
    elif isinstance(obj, dict):
        return {
            apply_func_on_object_recursively(key, f, apply_key_dict)
            if apply_key_dict
            else key: apply_func_on_object_recursively(value, f, apply_key_dict)
            for key, value in obj.items()
        }
    else:
        return f(obj)


async def apply_async_func_on_object_recursively(
    obj: Any, f: Callable[..., Awaitable[Any]], **kwargs
) -> Any:
    if isinstance(obj, list):
        return [await apply_async_func_on_object_recursively(value, f, **kwargs) for value in obj]

    if isinstance(obj, tuple):
        return tuple(  # pylint: disable=consider-using-generator
            [await apply_async_func_on_object_recursively(value, f, **kwargs) for value in obj]
        )

    if isinstance(obj, dict):
        keys = obj.keys()
        values = [
            await apply_async_func_on_object_recursively(v, f, **kwargs) for v in obj.values()
        ]
        return {key: values[num] for num, key in enumerate(keys)}

    return await f(obj, **kwargs)



from terality._terality.utils.config import TeralityConfig, TeralityCredentials


def has_valid_terality_config_file() -> bool:
    try:
        TeralityConfig.load()
        return True
    # TODO: finer-grained exception handling
    except Exception:
        return False


def has_valid_terality_credentials_file() -> bool:
    try:
        TeralityCredentials.load()
        return True
    # TODO: finer-grained exception handling
    except Exception:
        return False



import os
from typing import List, Tuple
from terality._terality.httptransport import HttpTransport
import inspect
from inspect import FrameInfo
import traceback
import pathlib

from common_client_scheduler.requests_responses import ClientErrorContext

from .constants import TERALITY_DISABLE_TELEMETRY_ENV_VAR


class ErrorReporter:
    """Reports errors (with context) to the Terality API.

    As a companion to Sentry error reporting, we add our own error reporter with a different focus:
    * we control our server-side sampling strategy
    * we can set our own limits on the metadata sent by this class

    This class is much less powerful than Sentry, and it's not obvious we'll keep it in the long run (this
    will depends on how our error reporting needs evolve). When adding metadata to errors, always consider
    adding them to Sentry events instead.
    """

    def __init__(self, http_transport: HttpTransport):
        self._http_transport = http_transport

    def report_exception(self, e: Exception) -> None:
        """Report a client exception to Terality servers.

        This method will retry to retrieve the code of the client module (whether running in a Python
        interpreter, a REPL, or a Jupyter notebook) that triggered the error. It does so by inspecting
        the current stack frame, and finding the first parent frame that's not inside the terality module.
        This is not foolproof (the terality module could call into a third party lib that calls back to
        the terality module) but should cover most of the actual case encountered in actual user scripts.

        Args:
            e: the exception to report. The exception traceback itself is ignored (the current stack frame
            is used instead, see above).

        See:
            report_exception_noexcept
        """
        if os.environ.get(TERALITY_DISABLE_TELEMETRY_ENV_VAR, "0") == "1":
            return

        tb = e.__traceback__  # pylint: disable=invalid-name
        if tb is None or not _originates_from_terality_module(traceback.extract_tb(tb)):
            return

        if _is_ipython():
            name = "ipython"
            code = _get_ipython_history()
        else:
            name, code = _get_calling_module_name_and_code()

        module_error_context = ClientErrorContext(
            code=code, filename=name, exception_str=f"{type(e).__name__}: {str(e)}"
        )
        self._http_transport.request("errors/contexts", payload=module_error_context)

    def report_exception_noexcept(self, e: Exception) -> None:
        """Report a client exception to Terality servers.

        This method will not raise any Exception (or subclass thereof).

        Args:
            e: the exception to report
        """
        try:
            self.report_exception(e)
        except Exception:
            pass  # silently ignore exceptions during error reporting


def _get_terality_module_filesystem_root():
    root_path = pathlib.Path(__file__).parent.parent
    assert (root_path / "version.py").exists()  # sanity check
    return root_path


def _originates_from_terality_module(stack_summary: traceback.StackSummary) -> bool:
    terality_module_root = _get_terality_module_filesystem_root()
    for frame in stack_summary:
        if str(terality_module_root) in frame.filename:
            return True
    return False


def _get_first_non_terality_module_path_from_stack(stack: List[FrameInfo]) -> str:
    terality_module_root = _get_terality_module_filesystem_root()

    first_frame_outside_terality_module = None
    for frame in stack:
        if str(terality_module_root) not in frame.filename:
            first_frame_outside_terality_module = frame
            break

    assert first_frame_outside_terality_module is not None
    return first_frame_outside_terality_module.filename


def _is_ipython() -> bool:
    """Return true if the current execution environment is a IPython kernel.

    This will return true when Terality is called from a Jupyter notebook.
    """
    try:
        # IPython injects the `get_ipython` method in the namespace
        get_ipython()  # type: ignore # noqa # pylint: disable=undefined-variable
        return True
    except NameError:
        return False


def _get_calling_module_name_and_code() -> Tuple[str, str]:
    """Return the code of the first module that's not `terality` in the current call stack.

    Raise:
        FileNotFoundError: if the code from the calling module can't be read.
    """
    current_stack = inspect.stack()
    calling_module_path = _get_first_non_terality_module_path_from_stack(current_stack)
    with open(calling_module_path) as f:
        calling_module_code = f.read()
    return calling_module_path, calling_module_code


def _get_ipython_history() -> str:
    """Return the content of the last 500 cells (or lines) executed by a IPython kernel.

    Only includes the user input, not the cell output.
    """
    # `get_ipython` will be defined in a IPython (used by a Jupyter notebook) context
    # as a safety check, only get the last 500 cells ran by the user
    input_cells = get_ipython().user_ns["_ih"][  # type: ignore # noqa # pylint: disable=undefined-variable
        -500:
    ]
    # cells can contain "\n", so add our own explicit delimiter between cells ("\n".join(...) would
    # lose the cell boundary information)
    return "\n\n##~ Cell delimiter ~##\n\n".join(input_cells)



from __future__ import annotations
import asyncio
import base64
from enum import Enum
import re
import zlib
import inspect
import json
import sys
from typing import Any, Dict, Optional, Set, Type, Union, List

from terality_serde.protobuf_wrapper import ProtobufWrapper
from terality_serde import TeralityDeserializationError

from . import TeralitySerializationError, all_external_types, external_base_classes, TypeSerde


_HAS_CACHED_PROPERTIES = sys.version_info[1] >= 8
if _HAS_CACHED_PROPERTIES:
    from functools import cached_property


# Subclasses of SerdeMixin can't be serialized in multiple threads in parallel
# (this mixin is not thread safe)
class SerdeMixin:
    subclasses: List[Type[SerdeMixin]] = []

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        cls.subclasses.append(cls)

    @property
    def class_name(self) -> str:
        return self.__class__.__name__

    @classmethod
    def _cached_properties_names(cls) -> Set[str]:
        # inspect.getmembers is expensive and dominate execution times if not cached
        if (
            not hasattr(cls, "_cached_properties_names_cache")
            or cls._cached_properties_names_cache is None  # type: ignore[attr-defined]
        ):
            cls._cached_properties_names_cache = {  # type: ignore[attr-defined]
                k for k, v in inspect.getmembers(cls) if isinstance(v, cached_property)
            } | {
                # betterproto internal fields
                "_unknown_fields",
                "_serialized_on_wire",
                "_group_map",
                "_group_current",
            }
        return cls._cached_properties_names_cache  # type: ignore[attr-defined]

    @property
    def dict(self) -> Dict:
        if _HAS_CACHED_PROPERTIES:
            dict_ = {
                k: v for k, v in self.__dict__.items() if k not in self._cached_properties_names()
            }
        else:
            dict_ = dict(self.__dict__)
        return dict_

    @classmethod
    def from_dict(cls, **kwargs):
        return cls(**kwargs)


class SerializableEnum(SerdeMixin, Enum):
    """An Enum that is also serializable.

    Usage:
    >>> class MySerializableEnum(SerializableEnum):
            VARIANT_1 = "VARIANT_1"
            VARIANT_2 = "VARIANT_2"
    """

    @property
    def dict(self) -> Dict:
        dict_ = {"name": self.name}
        return dict_

    @classmethod
    def from_dict(cls, name: str) -> SerializableEnum:  # type: ignore
        return cls[name]


class _SerdeConfig:
    """Store the known serialized and deserializer for various types, and apply them to objects.

    This class is immutable and a single instance can be reused.  For performance reasons, do not instantiate
    this class in tight loops.

    Note: the naming of this class implies that it's only configuration, but it actually contains business
    logic. Additionally, the naming of methods and attributes is all over the place ("serializers" are actually "encoders").
    Cleaning that up is TODO.
    """

    def __init__(self, all_external_types_):
        self._internal_type_attribute = "!terality:internal_type"
        self._external_type_attribute = "!terality:external_type"
        self._external_types = all_external_types_
        self._external_base_types = external_base_classes
        self._external_types_to_serializer = {ex.class_: ex for ex in all_external_types_}
        self._external_type_names_to_deserializer = {
            ex.class_name: ex for ex in self._external_types + self._external_base_types
        }
        self._internal_types_cached = None

    def _step_1_protobuf(self, obj: ProtobufWrapper):
        dict_ = {"protobuf": obj.serialize(), self._internal_type_attribute: obj.__class__.__name__}
        return dict_

    def _step_1_serde(self, obj: Any):
        dict_ = obj.dict
        dict_[self._internal_type_attribute] = obj.class_name
        return dict_

    def _step_2_serde(self, obj: Any):
        if isinstance(obj, dict) and all((isinstance(key, str) for key in obj.keys())):
            return obj

        external_type = self._external_types_to_serializer[type(obj)]
        dict_ = external_type.to_json(obj)
        dict_[self._external_type_attribute] = external_type.class_name
        return dict_

    def _step_3_serde(self, base_class, obj):
        dict_ = base_class.to_json(obj)
        dict_[self._external_type_attribute] = base_class.class_name
        return dict_

    def encode(self, obj: Any) -> Union[dict, Any]:
        """
        Encode an Python object to a JSON-serializable value.

        Objects of known types (types with a registered serializer) will be encoded to a dict.
        Objects of types that are already JSON-serializable and types without any registered serializer will be returned
        as-it.

        To fully encode an object to a JSON-serializable value, this function needs to be called recursively
        on any dicts or lists returned by this function (see `serialize`).
        """
        if isinstance(obj, ProtobufWrapper):
            proto = self._step_1_protobuf(obj)
            return proto

        if isinstance(obj, SerdeMixin):
            return self._step_1_serde(obj)

        type_ = type(obj)

        if type_ in self._external_types_to_serializer:
            return self._step_2_serde(obj)

        for base_class in self._external_base_types:
            if issubclass(type_, base_class.class_):
                return self._step_3_serde(base_class, obj)

        return obj

    def encode_recursively(self, obj):
        """Try to return a JSON-serializable representation of this object.

        Calls `encode` recursively. If no unknown types are encountered, the result is a JSON-serializable list, dict
        or scalar.

        Complex types will be encoded as dicts (see `encode`), then `encode` will be called recursively on the keys and
        values of the resulting dictionary (this allows encoding collections of arbitrary objects).

        The result of this function may not be JSON-serializable if this recursive process encounters types
        without any registered serializer.
        """
        encoded = self.encode(obj)

        if isinstance(encoded, list):
            return [self.encode_recursively(elt) for elt in encoded]
        if isinstance(encoded, dict):
            return {
                self.encode_recursively(k): self.encode_recursively(v) for k, v in encoded.items()
            }
        return encoded

    @property
    def _internal_types(self) -> Dict[str, Type[SerdeMixin]]:
        # Put in a property so that it gets computed at runtime at the last minute before it is needed
        # We do this because the problem of the subclasses registration is that it only happens if the subclass gets
        # imported in an __init__ somewhere, so we compute this function at the latest moment to give a chance for
        # the subclasses to have be imported
        if self._internal_types_cached is None:
            self._internal_types_cached = {
                internal.__name__: internal for internal in SerdeMixin.subclasses
            }
        return self._internal_types_cached

    @property
    def _protobuf_wrapper_types(self) -> Dict[str, Type[ProtobufWrapper]]:
        return {protobuf.__name__: protobuf for protobuf in ProtobufWrapper.subclasses}

    def decode(self, encoded: Dict[str, Any]):
        """Reverse of `encode`. Call this as `json.loads(object_hook=SerdeConfig().decode)`.

        Mutate `encoded`.
        """

        if self._internal_type_attribute in encoded:
            internal_type_name = encoded.pop(self._internal_type_attribute)

            # Detecting protobuf objects mixed with SerdeMixin
            if internal_type_name in self._protobuf_wrapper_types:
                protobuf_type = self._protobuf_wrapper_types[internal_type_name]
                return protobuf_type.parse(encoded["protobuf"])

            internal_type = self._internal_types[internal_type_name]
            try:
                return internal_type.from_dict(**encoded)
            except Exception as e:
                raise TeralityDeserializationError(
                    f"Can not instantiate an object of type {internal_type_name} with parameters {encoded}"
                ) from e

        if self._external_type_attribute in encoded:
            external_type_name = encoded.pop(self._external_type_attribute)
            deserializer = self._external_type_names_to_deserializer[external_type_name]
            try:
                return deserializer.from_json(**encoded)
            except Exception as e:
                raise TeralityDeserializationError(
                    f"Can not instantiate an object of type {external_type_name} with parameters {encoded}"
                ) from e

        return encoded


_SERDE_CONFIG = _SerdeConfig(all_external_types + [TypeSerde()])


# pylint: disable=invalid-name
def dumps(o: Any, compressed: bool = False, config: _SerdeConfig = _SERDE_CONFIG) -> str:
    o = config.encode_recursively(o)
    try:
        result = json.dumps(o)
    except TypeError as e:
        unserializable_type = _parse_json_type_error(e)
        if unserializable_type is not None:
            raise TeralitySerializationError(
                f"Can not serialize an object (or a collection or object containing such a type) of type'{unserializable_type}'\n{o}",
                unserializable_type=unserializable_type,
            ) from e
        raise TeralitySerializationError(
            f"Couldn't serialize the provided Python object: {str(e)}"
        ) from e
    if compressed:
        result = base64.encodebytes(zlib.compress(result.encode("utf-8"))).decode("utf-8")
    return result


# pylint: disable=invalid-name
def loads(s: str, compressed: bool = False, config: _SerdeConfig = _SERDE_CONFIG) -> Any:
    if compressed:
        s = zlib.decompress(base64.decodebytes(s.encode("utf-8"))).decode("utf-8")
    return json.loads(s, object_hook=config.decode)


def _parse_json_type_error(e: TypeError) -> Optional[str]:
    """Return the unserializable type that caused the JSON module to throw an error.

    The exception message should look like: "Object of type xxx is not JSON serializable".

    Returns None is the exception message could not be parsed.
    """
    if len(e.args) != 1 or not isinstance(e.args[0], str):
        return None
    message = e.args[0]
    match = re.compile(r"^Object of type (.*) is not JSON serializable$").match(message)
    if match is None:
        return None
    return match.group(1)


# pylint: disable=invalid-name
async def dumps_async(o: Any, compressed: bool = False) -> str:
    return await asyncio.get_event_loop().run_in_executor(None, dumps, o, compressed)


# pylint: disable=invalid-name
async def loads_async(
    s: str,
    compressed: bool = False,
) -> Any:
    return await asyncio.get_event_loop().run_in_executor(None, loads, s, compressed)



from dataclasses import dataclass, field
from functools import partial, partialmethod
import sys
from typing import Any, Dict, Set, Callable, Optional, Iterable, Tuple

# noinspection PyProtectedMember
from pandas._libs.properties import AxisProperty, CachedProperty

from . import call_pandas_function


# WARNING: `sys.gettrace` is considered a CPython implementation detail, so we check that it's available.
# See [documentation](https://docs.python.org/3.8/library/sys.html?highlight=gettrace#sys.gettrace).
# CTRacer is excluded because it is not a debugger but a code coverage tool.
_process_is_being_debugged: bool = (
    hasattr(sys, "gettrace")
    and sys.gettrace() is not None
    and sys.gettrace().__class__.__name__ != "CTracer"
)


class PandasDisplay:
    min_rows: int = 5
    max_rows: int = 20
    min_cols: int = 5
    max_cols: int = 20


class ClassMethod(type):
    """Only to be used for stuff that can't be done directly on yhe class"""

    _class_name: str
    _pandas_class: Any
    _additional_class_methods: Set[str] = {"from_pandas"}

    @classmethod
    def _attr_if_exists(cls, item: str):
        return getattr(cls._pandas_class, item, None)

    def __getattr__(cls, key: str) -> Callable:
        call = partial(call_pandas_function, cls._class_name, None, key)
        if key in cls._additional_class_methods or cls._attr_if_exists(key) is not None:
            return call
        raise AttributeError(f"type object '{cls._pandas_class.__name__}' has no attribute {key!r}")

    def __dir__(cls) -> Iterable[str]:
        return list(cls._additional_class_methods) + list(dir(cls._pandas_class))


class Struct(metaclass=ClassMethod):
    _pandas_class_instance: Any = None
    _additional_methods: Set[str] = {
        "to_pandas",
        "load",
        "save",
    }  # put in the mcs to get in the dir?
    _args_to_replace: Dict[str, Tuple[int, Callable]] = {}
    _accessors: Set[str] = set()  # put in the mcs to get in the dir?
    _indexers = {"iat", "at", "loc", "iloc"}  # put in the mcs to get in the dir?

    @classmethod
    def _is_property(cls, item: str) -> bool:
        property_ = cls._attr_if_exists(item)
        return isinstance(property_, (property, AxisProperty, CachedProperty))  # works for None too

    @classmethod
    def _is_method(cls, item: str) -> bool:
        method = cls._attr_if_exists(item)
        return callable(method)  # works for None too

    @classmethod
    def _call(cls, accessor, func_name, *args, **kwargs):
        return call_pandas_function(cls._class_name, accessor, func_name, *args, **kwargs)

    def _call_method(self, accessor: Optional[str], func_name: str, *args, **kwargs):
        if func_name in self._args_to_replace:
            arg_num, callable_ = self._args_to_replace[func_name]
            args = tuple(
                list(args[:arg_num]) + [callable_(args[arg_num])] + list(args[arg_num + 1 :])
            )
        return self._call(accessor, func_name, self, *args, **kwargs)

    def _on_missing_attribute(self, item: str):
        raise AttributeError(f"'{type(self)}' has no attribute {item!r}")

    ###################################################################################################################
    # Creation/upgrade

    @classmethod
    def _new(cls, id_: str, cached_attributes=None):
        # Workaround to avoid being intercepted
        struct = object.__new__(cls)
        struct._id = id_  # pylint: disable=attribute-defined-outside-init
        struct._cached_attributes = (  # pylint: disable=attribute-defined-outside-init
            {} if cached_attributes is None else cached_attributes
        )
        return struct

    def _mutate(self, other: "Struct") -> None:
        self._id = other._id  # type: ignore  # pylint: disable=attribute-defined-outside-init
        self._cached_attributes = (  # pylint: disable=attribute-defined-outside-init
            other._cached_attributes  # type: ignore
        )

    ###################################################################################################################
    # Magic methods

    def __new__(cls, *args, **kwargs):
        return cls._call(None, "__init__", *args, **kwargs)

    def __getattr__(self, item: str):  # pylint: disable=too-many-return-statements
        """Intercepting all not-magic methods"""
        if item.startswith("_") and item not in ("_repr_html_"):
            # _repr_html_ is a supported API call that returns an HTML representation of the struct.
            # All other IPython representation methods (such as _repr_svg or _ipython_display_) are currently
            # not supported and should raise an AttributeError client-side.
            # This will cause IPython to default to a supported representation, without needless requests to
            # the Terality API.
            return object.__getattribute__(self, item)
        if item in self._cached_attributes:  # type: ignore
            return self._cached_attributes[item]  # type: ignore
        if item in self._accessors:
            return NameSpace(self, item)
        if item in self._indexers:
            return Indexer(self, item)
        if item in self._additional_methods:
            return partial(self._call_method, None, item)
        if self._is_property(item):
            return self._call_method(None, item)
        if self._is_method(item):
            return partial(self._call_method, None, item)
        return self._on_missing_attribute(item)

    def __setattr__(self, name, value):  # pylint: disable=inconsistent-return-statements
        if name.startswith("_"):
            object.__setattr__(self, name, value)
        else:
            return self._call_method("__setattr__", name, value=value)

    def __str__(self):
        # NOTE: When debugging, we avoid network calls and fall back to a bare bones representation.
        if _process_is_being_debugged:
            _id = getattr(self, "_id", None)
            return f"terality.{self.__class__.__name__}(_id={_id})"
        return self._call_method(None, "__str__")

    def __repr__(self):
        return self.__str__()

    def __dir__(self) -> Iterable[str]:
        # WARNING: When the process is being debugged, we do not want to list any dynamic properties.
        # Why? Because the debugger's variable view uses `dir` to list a variable's children, then evaluates them.
        # Listing dynamic properties would cause spurious calls to the server.
        if _process_is_being_debugged:
            members = [
                key
                for key in self.__dict__.keys() | type(self).__dict__.keys()
                if key.startswith("_") and not key.startswith("__")
            ]
            return members

        members = list(self._additional_methods)
        if self._pandas_class_instance is not None:
            members += list(dir(self._pandas_class_instance))
        return members

    def __len__(self):
        return self._call_method(None, "__len__")


# Add magic methods (otherwise they won't be itercepted) programmatically rather than by hand
# ------------------------------


def _forward_to_backend(lhs: Struct, func_name: str, *args, **kwargs):
    return lhs._call_method(None, func_name, *args, **kwargs)


def _set_magics(magic_method_names):
    for magic_method_name in magic_method_names:
        full_magic_method_name = f"__{magic_method_name}__"
        operator = partialmethod(_forward_to_backend, full_magic_method_name)
        setattr(Struct, full_magic_method_name, operator)


_access = ["getitem", "setitem", "delitem"]
_comparisons = ["lt", "le", "eq", "ne", "ge", "gt"]
_others = ["neg", "pos", "invert", "abs", "len"]

_arithmetic = [
    "add",
    "sub",
    "mul",
    "matmul",
    "truediv",
    "floordiv",
    "mod",
    "divmod",
    "pow",
]
_boolean = [
    "and",
    "xor",
    "or",
]
_other = [
    "lshift",
    "rshift",
]
_with_reverse = _other + _arithmetic + _boolean
_double = [f"{prefix}{function_name}" for function_name in _with_reverse for prefix in ["", "r"]]


_set_magics(_access + _comparisons + _others + _double)


# Other stuff
# ------------------------------


@dataclass
class NameSpace:
    _obj: Struct
    _name: str

    def __getattr__(self, item: str):
        if item.startswith("_"):
            return object.__getattribute__(self, item)
        pd_class = self._obj.__class__._attr_if_exists(self._name)
        pd_method = getattr(pd_class, item)
        if isinstance(pd_method, property):
            # noinspection PyProtectedMember
            return self._obj._call_method(self._name, item)
        # noinspection PyProtectedMember
        return partial(self._obj._call_method, self._name, item)


@dataclass
class Indexer:
    _obj: Struct
    _name: str

    def __getitem__(self, item):
        # noinspection PyProtectedMember
        return self._obj._call_method(self._name, "__getitem__", item)

    def __setitem__(self, key, value):
        # noinspection PyProtectedMember
        self._obj._call_method(self._name, "__setitem__", key, value)


@dataclass
class StructIterator:
    _struct: Struct
    _pos: int = 0
    _buffer: list = field(default_factory=list)
    _buffer_start: int = 0

    def __iter__(self):
        # Must return itself, as per the [documentation](https://docs.python.org/3.8/library/stdtypes.html#typeiter).
        return self

    @property
    def _buffer_stop(self):
        return self._buffer_start + len(self._buffer)

    def __next__(self) -> Any:
        if self._pos >= self._buffer_stop:
            self._buffer_start = self._pos
            self._buffer = self._struct.get_range_auto(self._buffer_start)
            if len(self._buffer) == 0:
                raise StopIteration()
        value = self._buffer[self._pos - self._buffer_start]
        self._pos += 1
        return value



import re
from abc import ABC, abstractmethod
import base64
from dataclasses import dataclass
from datetime import datetime, time, timedelta, date
from io import BytesIO
from pathlib import Path, PosixPath
from re import Pattern
from typing import Any, List, Optional, Set, Type, Union

import numpy as np
import pandas as pd
from pandas.tseries.offsets import (
    Nano,
    Micro,
    Milli,
    Easter,
    FY5253,
    BusinessHour,
    CustomBusinessDay,
    Second,
    Minute,
    Hour,
    Day,
    BusinessDay,
    Week,
    SemiMonthBegin,
    SemiMonthEnd,
    MonthBegin,
    MonthEnd,
    BMonthEnd,
    BMonthBegin,
    WeekOfMonth,
    LastWeekOfMonth,
    BQuarterBegin,
    BQuarterEnd,
    QuarterBegin,
    QuarterEnd,
    YearBegin,
    YearEnd,
    BYearBegin,
    BYearEnd,
)

from dateutil.parser import ParserError

# noinspection PyProtectedMember
from pandas._libs import missing, OutOfBoundsDatetime, tslibs
import pyarrow


@dataclass  # type: ignore
class ExternalTypeSerializer(ABC):
    class_: type

    @property
    def class_name(self) -> str:
        return f"{self.class_.__module__}.{self.class_.__name__}"

    @abstractmethod
    def to_json(self, value: Any) -> dict:
        pass

    @abstractmethod
    def from_json(self, **kwargs):
        pass


@dataclass
class TupleSerde(ExternalTypeSerializer):
    class_: type = tuple

    def to_json(self, value: tuple) -> dict:
        return {"as_list": list(value)}

    def from_json(self, as_list: list) -> tuple:  # type: ignore
        return tuple(as_list)


@dataclass
class OffsetsGenericSerde(ExternalTypeSerializer):
    class_: type
    useful_attributes: List[str]

    def to_json(self, value) -> dict:
        return {attribute: getattr(value, attribute) for attribute in self.useful_attributes}

    def from_json(self, **kwargs):
        return self.class_(**kwargs)


@dataclass
class DictSerde(ExternalTypeSerializer):
    """
    Serialize dict with non-strings keys as json only handle strings keys.
    """

    class_: type = dict

    def to_json(self, value: dict) -> dict:
        return {"keys": list(value.keys()), "values": list(value.values())}

    def from_json(self, keys: list, values: list) -> dict:  # type: ignore
        assert len(keys) == len(values)
        return dict(zip(keys, values))


@dataclass
class RangeSerde(ExternalTypeSerializer):
    class_: type = range

    def to_json(self, value: range) -> dict:
        return {"as_list": [value.start, value.stop, value.step]}

    def from_json(self, as_list: list) -> range:  # type: ignore
        return range(*as_list)


@dataclass
class SliceSerde(ExternalTypeSerializer):
    class_: type = slice

    def to_json(self, value: slice) -> dict:
        return {"as_list": [value.start, value.stop, value.step]}

    def from_json(self, as_list: list) -> slice:  # type: ignore
        return slice(*as_list)


@dataclass
class TimeSerde(ExternalTypeSerializer):
    class_: type = time

    def to_json(self, value: time) -> dict:
        return {"isoformat": value.isoformat()}

    def from_json(self, isoformat: str) -> time:  # type: ignore
        return time.fromisoformat(isoformat)


@dataclass
class DateSerde(ExternalTypeSerializer):
    class_: type = date

    def to_json(self, value: date) -> dict:
        return {"isoformat": value.isoformat()}

    def from_json(self, isoformat: str) -> date:  # type: ignore
        return date.fromisoformat(isoformat)


@dataclass
class DateTimeSerde(ExternalTypeSerializer):
    class_: type = datetime

    def to_json(self, value: datetime) -> dict:
        return {"isoformat": value.isoformat()}

    def from_json(self, isoformat: str) -> datetime:  # type: ignore
        return datetime.fromisoformat(isoformat)


@dataclass
class TimeStampSerde(ExternalTypeSerializer):
    """
    Pandas does not handle the timezones in isoformat, so we have to manually
    remove/put back timezones in serde...
    """

    class_: type = pd.Timestamp

    def to_json(self, value: pd.Timestamp) -> dict:
        timezone = None if value.tz is None else str(value.tz)
        if timezone is not None:
            value = value.astimezone(None)
        return {"isoformat": value.value, "timezone": timezone}

    def from_json(self, isoformat: int, timezone: Optional[str]) -> pd.Timestamp:  # type: ignore
        timestamp = pd.to_datetime(isoformat)
        # timestamp = pd.Timestamp.fromisoformat(isoformat)
        if timezone is not None:
            timestamp = timestamp.tz_localize("UTC").astimezone(timezone)
        return timestamp


@dataclass
class PeriodSerde(ExternalTypeSerializer):
    """
    We have a dtype to handle array of Period, but we need this serde to pass scalar as function parameter.
    This serde is tested in df_setitem.
    """

    class_: type = pd.Period

    def to_json(self, value: pd.Period) -> dict:
        isoformat = value.to_timestamp().value
        return {"isoformat": isoformat, "freq": value.freq.name}

    def from_json(self, isoformat: int, freq: str) -> pd.Period:  # type: ignore
        timestamp = pd.to_datetime(isoformat)
        return pd.Period(
            year=timestamp.year,
            month=timestamp.month,
            day=timestamp.day,
            hour=timestamp.hour,
            minute=timestamp.minute,
            second=timestamp.second,
            freq=freq,
        )


@dataclass
class IntervalSerde(ExternalTypeSerializer):
    """
    We have a dtype to handle array of Interval, but we need this serde to pass scalar as function parameter.
    This serde is tested in df_setitem.
    """

    class_: type = pd.Interval

    # TODO Handle non-builtin scalar for left/right
    def to_json(self, value: pd.Interval) -> dict:
        return {"left": value.left, "right": value.right, "closed": value.closed}

    def from_json(self, left, right, closed: str) -> pd.Interval:  # type: ignore
        return pd.Interval(left=left, right=right, closed=closed)


@dataclass
class TimedeltaSerde(ExternalTypeSerializer):
    class_: type = pd.Timedelta

    def to_json(self, value: pd.Timedelta) -> dict:
        return {"value_ns": value.value}

    def from_json(self, value_ns: int) -> Union[pd.Timedelta, type(pd.NaT)]:  # type: ignore
        return pd.Timedelta(value=value_ns, unit="ns")


@dataclass
class PyTimedeltaSerde(ExternalTypeSerializer):
    class_: type = timedelta

    def to_json(self, value: timedelta) -> dict:
        return {"days": value.days, "seconds": value.seconds, "microseconds": value.microseconds}

    def from_json(self, days: int, seconds: int, microseconds: int) -> timedelta:  # type: ignore
        return timedelta(days=days, seconds=seconds, microseconds=microseconds)


@dataclass
class PathSerde(ExternalTypeSerializer):
    class_: type = PosixPath

    def to_json(self, value: Path) -> dict:
        return {"path_as_str": str(value)}

    def from_json(self, path_as_str: str) -> Path:  # type: ignore
        return Path(path_as_str)


@dataclass
class BytesSerde(ExternalTypeSerializer):
    class_: type = bytes

    def to_json(self, value: bytes) -> dict:
        # ascii85 may be more space-efficient, but it includes many characters that would need to be escaped
        # when sent over HTTP. Overall, let's stick to something that doesn't include weird characters,
        # and we'll revisit this if performance becomes an issue.
        return {"bytes_base64": base64.b64encode(value).decode("utf-8")}

    def from_json(self, bytes_base64: bytes) -> bytes:  # type: ignore
        return base64.b64decode(bytes_base64)


@dataclass
class NumpyDType(ExternalTypeSerializer):
    class_: type = np.dtype

    def to_json(self, value: np.dtype) -> dict:
        return {"dtype_name": value.name}

    def from_json(self, dtype_name: str) -> np.dtype:  # type: ignore
        return np.dtype(dtype_name)


@dataclass
class NumpyArray(ExternalTypeSerializer):
    class_: type = np.ndarray

    def to_json(self, value: np.ndarray) -> dict:
        return {
            "buffer": base64.b64encode(value.tobytes()).decode("ascii"),
            "dtype": str(value.dtype),
        }

    def from_json(self, buffer: str, dtype: str) -> np.number:  # type: ignore
        return np.frombuffer(base64.b64decode(buffer.encode("ascii")), dtype=dtype)


@dataclass
class NumpyScalarSerde(ExternalTypeSerializer):
    def to_json(self, value: np.number) -> dict:
        return {"value": value.item()}

    def from_json(self, value: str) -> np.number:  # type: ignore
        return self.class_(value)


numpy_scalars = [
    NumpyScalarSerde(np_scalar_type)
    for np_scalar_type in [
        np.bool_,
        np.uint8,
        np.uint16,
        np.uint32,
        np.uint64,
        np.int8,
        np.int16,
        np.int32,
        np.int64,
        np.float64,
    ]
]


@dataclass
class NumpyF32Serde(ExternalTypeSerializer):
    class_: type = np.float32

    def to_json(self, value: np.float32) -> dict:
        bytes_io = BytesIO()
        # noinspection PyTypeChecker
        np.save(bytes_io, value)
        o_bytes = bytes_io.getvalue()
        o_str = base64.b64encode(o_bytes).decode()
        return {"float_as_array": o_str}

    def from_json(self, float_as_array: str) -> np.float32:  # type: ignore
        bytes_io = BytesIO(base64.b64decode(float_as_array.encode()))
        # Numpy signature does  not list all variants
        # noinspection PyTypeChecker
        return np.load(bytes_io).flatten()[0]


@dataclass
class NumpyDateTime64Serde(ExternalTypeSerializer):
    class_: type = np.datetime64

    def to_json(self, value: np.datetime64) -> dict:
        # This is horrible, best bet seems to go through pandas to manage to serde them:
        # https://stackoverflow.com/questions/13703720/converting-between-datetime-timestamp-and-datetime64
        return {"datetime_as_isoformat": pd.Timestamp(value).isoformat()}

    def from_json(self, datetime_as_isoformat: str) -> np.datetime64:  # type: ignore
        return np.datetime64(datetime_as_isoformat)


@dataclass
class NumpyBusDayCalendarSerde(ExternalTypeSerializer):
    class_: type = np.busdaycalendar

    def to_json(self, value: np.busdaycalendar) -> dict:
        return {"weekmask": value.weekmask, "holidays": value.holidays}

    def from_json(  # type: ignore
        self, weekmask: Union[str, List[bool]], holidays: List[np.datetime64]
    ) -> np.busdaycalendar:
        return np.busdaycalendar(weekmask, holidays)


@dataclass
class PandasMissingSerde(ExternalTypeSerializer):
    value: Any

    def to_json(self, value: np.number) -> dict:
        return {}

    def from_json(self) -> np.number:  # type: ignore
        return self.value


@dataclass
class PandasNASerde(PandasMissingSerde):
    class_: type = missing.NAType  # pylint: disable=c-extension-no-member
    value: Any = pd.NA


@dataclass
class PandasNaTSerde(PandasMissingSerde):
    class_: type = tslibs.NaTType
    value: Any = pd.NaT


@dataclass
class PyarrowDataType(ExternalTypeSerializer):
    class_: type = pyarrow.DataType

    _mapping = {
        data_type.id: data_type
        for data_type in [
            pyarrow.null(),
            pyarrow.string(),
            pyarrow.date32(),
            pyarrow.date64(),
        ]
    }

    def to_json(self, value: pyarrow.DataType) -> dict:
        return {"value": value.id}

    def from_json(self, value: str) -> type:  # type: ignore
        return self._mapping[value]


@dataclass
class PandasDType(ExternalTypeSerializer):
    class_: type = pd.api.extensions.ExtensionDtype

    def to_json(self, value: pd.api.extensions.ExtensionDtype) -> dict:
        return {"dtype_name": value.name}

    def from_json(self, dtype_name: str) -> np.dtype:  # type: ignore
        return pd.api.types.pandas_dtype(dtype_name)


@dataclass
class PandasCategoricalSerde(ExternalTypeSerializer):
    class_: type = pd.Categorical

    def to_json(self, value: pd.Categorical) -> dict:
        return {
            "codes": value.codes.tolist(),
            "categories": value.categories.tolist(),
            "ordered": value.ordered,
        }

    def from_json(self, codes: str, categories, ordered: bool) -> pd.Categorical:  # type: ignore
        return pd.Categorical.from_codes(codes=codes, categories=categories, ordered=ordered)


@dataclass
class PandasIntervalSerde(ExternalTypeSerializer):
    class_: type = pd.Interval

    def to_json(self, value: pd.Interval) -> dict:
        return {"left": value.left, "right": value.right, "closed": value.closed}

    def from_json(self, left: Any, right: Any, closed: str) -> pd.Interval:  # type: ignore
        return pd.Interval(left=left, right=right, closed=closed)


@dataclass
class RegexSerde(ExternalTypeSerializer):
    class_: type = Pattern

    def to_json(self, regex: Pattern) -> dict:
        return {"pattern": regex.pattern}

    def from_json(self, pattern: str) -> Pattern:  # type: ignore
        return re.compile(pattern)


external_base_classes = [NumpyDType(), PandasDType()]
# _pandas_dtype_classes = [
#     pd.StringDtype,
#     pd.CategoricalDtype,
#     pd.BooleanDtype,
#     pd.DatetimeTZDtype,
#     pd.Int8Dtype,
#     pd.Int16Dtype,
#     pd.Int32Dtype,
#     pd.Int64Dtype,
#     pd.UInt8Dtype,
#     pd.UInt16Dtype,
#     pd.UInt32Dtype,
#     pd.UInt64Dtype,
# ]
# pandas_dtypes = [PandasDType(pandas_dtype) for pandas_dtype in _pandas_dtype_classes]


_whitelisted_types = {
    bool,
    int,
    float,
    str,
    object,
    complex,
    np.float64,
    np.float32,
    np.uint8,
    np.uint16,
    np.uint32,
    np.uint64,
    np.int8,
    np.int16,
    np.int32,
    np.int64,
    AssertionError,
    AttributeError,
    IndexError,
    OutOfBoundsDatetime,
    ParserError,
    ValueError,
    TypeError,
    ZeroDivisionError,
}


@dataclass
class TypeSerde(ExternalTypeSerializer):
    class_: type = type

    def __init__(self, other_types: Optional[Set[Type]] = None):
        if other_types is None:
            self._whitelisted_types = _whitelisted_types
        else:
            self._whitelisted_types = _whitelisted_types | other_types

    @staticmethod
    def _fullname(type_: type) -> str:
        return f"{type_.__module__}.{type_.__qualname__}"

    def to_json(self, value: type) -> dict:
        return {"value": self._fullname(value)}

    def from_json(self, value: str) -> type:  # type: ignore
        return {self._fullname(class_): class_ for class_ in self._whitelisted_types}[value]


scalar_types: List[ExternalTypeSerializer] = [
    DateSerde(),
    DateTimeSerde(),
    TimeStampSerde(),
    PeriodSerde(),
    IntervalSerde(),
    TimeSerde(),
    # NumpyDType(),
    NumpyF32Serde(),
    NumpyDateTime64Serde(),
    NumpyBusDayCalendarSerde(),
    NumpyArray(),
    PandasNASerde(),
    PandasNaTSerde(),
    TimedeltaSerde(),
    PyTimedeltaSerde(),
    PandasCategoricalSerde(),
    PandasIntervalSerde(),
    RegexSerde(),
]
scalar_types += numpy_scalars

tseries_offset_types: List[ExternalTypeSerializer] = [
    OffsetsGenericSerde(Second, ["n"]),
    OffsetsGenericSerde(Minute, ["n"]),
    OffsetsGenericSerde(BusinessHour, ["n", "normalize", "start", "end"]),
    OffsetsGenericSerde(Hour, ["n"]),
    OffsetsGenericSerde(BusinessDay, ["n"]),
    OffsetsGenericSerde(CustomBusinessDay, ["n", "normalize", "calendar", "offset"]),
    OffsetsGenericSerde(Day, ["n"]),
    OffsetsGenericSerde(Week, ["n", "weekday"]),
    OffsetsGenericSerde(SemiMonthBegin, ["n", "day_of_month", "normalize"]),
    OffsetsGenericSerde(SemiMonthEnd, ["n", "day_of_month", "normalize"]),
    OffsetsGenericSerde(MonthBegin, ["n"]),
    OffsetsGenericSerde(MonthEnd, ["n"]),
    OffsetsGenericSerde(BMonthEnd, ["n"]),
    OffsetsGenericSerde(BMonthBegin, ["n"]),
    OffsetsGenericSerde(WeekOfMonth, ["n", "week", "weekday"]),
    OffsetsGenericSerde(LastWeekOfMonth, ["n", "weekday"]),
    OffsetsGenericSerde(BQuarterBegin, ["n", "startingMonth"]),
    OffsetsGenericSerde(BQuarterEnd, ["n", "startingMonth"]),
    OffsetsGenericSerde(QuarterBegin, ["n", "startingMonth"]),
    OffsetsGenericSerde(QuarterEnd, ["n", "startingMonth"]),
    OffsetsGenericSerde(BYearBegin, ["n", "month"]),
    OffsetsGenericSerde(BYearEnd, ["n", "month"]),
    OffsetsGenericSerde(YearBegin, ["n", "month"]),
    OffsetsGenericSerde(YearEnd, ["n", "month"]),
    OffsetsGenericSerde(FY5253, ["n", "weekday", "startingMonth", "variation"]),
    OffsetsGenericSerde(Easter, ["n"]),
    OffsetsGenericSerde(Milli, ["n"]),
    OffsetsGenericSerde(Micro, ["n"]),
    OffsetsGenericSerde(Nano, ["n"]),
]
scalar_types += tseries_offset_types

_other_types = [
    DictSerde(),
    PyarrowDataType(),
    TupleSerde(),
    SliceSerde(),
    RangeSerde(),
    PathSerde(),
    BytesSerde(),
]

all_external_types = scalar_types + _other_types



from pandas.core import groupby
from terality_serde import StructType

from . import ClassMethod, Struct


class ClassMethodGroupBySeries(ClassMethod):
    _class_name: str = StructType.SERIES_GROUPBY
    # noinspection PyUnresolvedReferences
    _pandas_class = groupby.SeriesGroupBy


class SeriesGroupBy(Struct, metaclass=ClassMethodGroupBySeries):
    # noinspection PyUnresolvedReferences
    _pandas_class_instance = groupby.SeriesGroupBy
    _additional_methods = Struct._additional_methods | {"sum", "mean"}


class ClassMethodGroupByDF(ClassMethod):
    _class_name: str = StructType.DATAFRAME_GROUPBY
    # noinspection PyUnresolvedReferences
    _pandas_class = groupby.DataFrameGroupBy


class DataFrameGroupBy(Struct, metaclass=ClassMethodGroupByDF):
    # noinspection PyUnresolvedReferences
    _pandas_class_instance = groupby.DataFrameGroupBy
    _additional_methods = Struct._additional_methods | {"sum", "mean"}

    def _on_missing_attribute(self, item: str):
        return self._call_method(None, "df_groupby_attribute_access", item)



from functools import partial
from terality._terality.utils.config import TeralityConfig
from typing import Dict, Optional
from uuid import uuid4
import inspect

import pandas as pd

from common_client_scheduler import UploadRequest
from terality_serde import StructType

from .. import upload_local_files, upload_s3_files, global_client
from . import call_pandas_function
from ..utils.azure import parse_azure_filesystem, test_for_azure_libs
from ...exceptions import TeralityClientError


def _make_upload(path: str, storage_options: Optional[Dict] = None) -> UploadRequest:
    config = TeralityConfig.load(fallback_to_defaults=True)
    if config.skip_transfers:
        transfer_id = ""
        aws_region = None
    else:
        if path.startswith("s3://"):
            parts = path[len("s3://") :].split("/", 1)
            if len(parts) != 2:
                raise ValueError(
                    f"Invalid S3 path, expected format: 's3://bucket/prefix' (prefix may be the empty string), got: '{path}'."
                )
            transfer_id = upload_s3_files(s3_bucket=parts[0], s3_key_prefix=parts[1])
            aws_region = global_client().get_upload_config().default_aws_region
        elif path.startswith("abfs://") or path.startswith("az://"):
            test_for_azure_libs()
            from ..data_transfers.azure import upload_azure_storage_files

            storage_account_name, container, folder = parse_azure_filesystem(path, storage_options)
            transfer_id = upload_azure_storage_files(storage_account_name, container, folder)
            aws_region = global_client().get_upload_config().default_aws_region
        elif path.startswith("gs://"):
            raise TeralityClientError(
                "Currently, Terality does not support GCP storage. Please use AWS S3 or Azure paths."
            )
        else:
            transfer_id = str(uuid4())
            upload_local_files(path, transfer_id, global_client().get_aws_credentials())
            aws_region = None
    return UploadRequest(path=path, transfer_id=transfer_id, aws_region=aws_region)


def _treat_read_job(method_name, *args, **kwargs):
    """Special job to intercept file arguments and upload them to Terality for pd.read_xxx() jobs"""
    storage_options = kwargs.get("storage_options")
    path_parameter_name = _read_top_level_functions_to_path_parameter[method_name]
    if path_parameter_name in kwargs:
        kwargs[path_parameter_name] = _make_upload(kwargs[path_parameter_name], storage_options)
    else:
        if len(args) == 0:
            raise TypeError(
                f"{method_name}() missing 1 required positional argument: '{path_parameter_name}'"
            )
        path, *others = args
        args = [_make_upload(path, storage_options)] + others
    return call_pandas_function(StructType.TOP_LEVEL, None, method_name, *args, **kwargs)


# pandas top level functions (e.g defined as pd.SomeFunc) that Terality reimplements
# we also need to specify each path parameter name as we handle it similarly for all read methods
_read_top_level_functions_to_path_parameter = {
    "read_parquet": "path",
    "read_csv": "filepath_or_buffer",
    "read_excel": "io",
}

# Even if for now the `read_xxx` function are the only one to be reimplemented by Terality,
# they have a very specific path in the code.
# Should we add any new top level function in the future, it can be implemented very differently
# than a "read_xxx" function.
# Make this clear by already having two concepts: generic "top level functions" whose only instances
# for now are the read functions.
top_level_functions = list(_read_top_level_functions_to_path_parameter)


def get_dynamic_top_level_attribute(attribute: str):
    """Return the Terality or pandas module attribute with the given name.

    This function is intended to be the implementation of `__getattr__` (not `__getattribute__`) for the
    Terality module.

    When a user tries to access an attribute on the `terality` module that's not exported through the usual means,
    such as:

    >>> import terality as pd
    >>> df = pd.Something()

    then we use `_get_top_level_attribute` to dynamically resolve the `pd.Something` name.

    Note that data structures such as `pd.DataFrame` are exported in the `__init__` file of the Terality module
    and as such are never resolved through this function. Python only calls `__getattr__` when it can't resolve
    a name otherwise.

    The name resolution logic we use is:
    0. check that the pandas module defines the name. If not, raise an AttributeError
    1. intercept functions that need special client processing (such as read_csv, read_parquet...)
    2. if it's a class in pandas (`pd.Timestamp`), then return the pandas class
    3. if it's a module in pandas (`pd.tseries`), then return the pandas module
    4. otherwise, assume that it's a function, and send it to the Terality API

    In the future, we may implement an allowlist of functions to be locally executed by pandas (it could
    be sent by the server at session creation for instance).
    """
    # Check if the attribute is defined by the pandas library. If not, this line will raise an AttributeError.
    pd_attr = getattr(pd, attribute)

    # Special cases for NaT and NA
    if attribute == "NaT":
        return pd.NaT
    if attribute == "NA":
        return pd.NA

    # Intercept functions that need special handling (for now, only the "read_xxx" family of functions).
    if attribute in _read_top_level_functions_to_path_parameter:
        return partial(_treat_read_job, attribute)

    # Send class and module calls to the pandas module
    if inspect.isclass(pd_attr) or inspect.ismodule(pd_attr):
        return pd_attr

    # Send anything else to the Terality API
    return partial(call_pandas_function, StructType.TOP_LEVEL, None, attribute)



config_not_found = "Terality configuration file not found."  # pylint: disable=invalid-name


# pylint: disable=invalid-name
config_helper = """
- If you don't have a Terality account yet, go to https://terality.com/ to sign up!
- If you have a Terality account and received your credentials, call `terality.configure(user_id, user_password)` to set up your configuration.
- If you have moved the configuration somewhere else, specify its location using the `TERALITY_HOME` environment variable.
"""



"""Implements support for anonymous users: temporary, short-lived accounts that can be used to quickly evaluate Terality.

For any regular usage, creating a full-fledged Terality account is required.
"""

import datetime as dt

from common_client_scheduler.requests_responses import CreateAnonymousUserResponse

from terality._terality.client import anonymous_client
from terality._terality.globals import load_global_client
from terality._terality.utils.account import (
    has_valid_terality_config_file,
    has_valid_terality_credentials_file,
)
from terality._terality.utils.config import TeralityConfig, TeralityCredentials


def create_anonymous_user():
    if has_valid_terality_config_file() or has_valid_terality_credentials_file():
        # Note that we don't detect if the previous configuration is for a quickstart account or not.
        # This means that a user that used the quickstart method on this laptop then created a real account
        # with `terality account create` will be greeted with a "configuration files already exist" warning.
        print("Terality is already configured on this machine, nothing to do.")
        return

    client = anonymous_client()
    res = client.poll_for_answer("anonymous_users", payload={})
    if not isinstance(res, CreateAnonymousUserResponse):
        raise ValueError(
            f"Unexpected server answer: expected type 'CreateAnonymousUserResponse', got '{type(res)}'"
        )
    user_id = res.user_id
    api_key = res.api_key
    expires_at = res.expires_at
    _write_terality_configuration(user_id, api_key)

    print(
        f"Terality is now initialized on this machine. This is a temporary demonstration account that will only remain valid for {_format_account_duration_in_minutes(expires_at)} minutes. To create a full account, please go to https://app.terality.com (it's free!)."
    )


def _write_terality_configuration(email: str, api_key: str):
    credentials = TeralityCredentials(user_id=email, user_password=api_key)
    credentials.save()
    config = TeralityConfig()
    config.save()

    load_global_client()


def _format_account_duration_in_minutes(expires_at: dt.datetime) -> str:
    return str(round(max((expires_at - dt.datetime.utcnow()).total_seconds() // 60, 0)))



"""Utilities for Azure integration.

This module does not assume that the Azure Python SDK is installed (no Azure top-level import) and
can be safely imported without testing for the presence of the SDK.
"""

import os
from typing import Dict, Optional, Tuple


def test_for_azure_libs() -> bool:
    """Emit a warning if the Azure Python SDK is not installed.

    Return:
        true if the Azure Python SDK is installed.
    """
    try:
        return True
    except ImportError as e:
        import warnings

        warnings.warn(
            f"Azure Python SDK not found ({str(e)}). The SDK is required for the Azure integration. You can install the Azure extras with: pip install terality[azure]"
        )
        return False


def parse_azure_filesystem(
    fs_url: str, storage_options: Optional[Dict] = None
) -> Tuple[str, str, str]:
    """Return a tuple (storage account name, container name, blob name) from a Azure Datalake Gen2 filesystem URL.

    Mimics the implementation from adlfs (https://github.com/dask/adlfs), supporting only Gen2 storage.
    Pattern: "abfs://{container}/{path}", account_name in storage_options.
    Unlike abfs, for simplicy, we treat the path as either a folder name, or a file name (no support for globs)
    """
    if storage_options is None:
        storage_options = {}
    storage_account_name = storage_options.get("account_name")
    if storage_account_name is None:
        storage_account_name = os.environ.get("AZURE_STORAGE_ACCOUNT_NAME")
    if storage_account_name is None:
        raise ValueError(
            "Azure storage account name not set. Set the `storage_options` parameter to a dictionary containing the key `account_name`."
        )
    if fs_url.startswith("abfs://"):
        container, path = fs_url[len("abfs://") :].split("/", maxsplit=1)
    elif fs_url.startswith("az://"):
        container, path = fs_url[len("az://") :].split("/", maxsplit=1)
    else:
        raise ValueError(f"Unsupported filesystem scheme: '{fs_url}'")
    return storage_account_name, container, path



from __future__ import annotations
import time
from typing import Any, Optional

from terality.exceptions import TeralityError
from terality._terality.errorreporting import ErrorReporter

from common_client_scheduler.config import SessionInfo
from common_client_scheduler.requests_responses import (
    AwsCredentials,
    CreateSessionResponse,
    DataTransferResponse,
    PendingComputationResponse,
)

from .utils.config import TeralityConfig, TeralityCredentials
from .httptransport import AuthCredentials, HttpTransport
from .utils import logger


class UnconfiguredTeralityClientError(TeralityError):
    pass


class SessionStateError(TeralityError):
    pass


class TeralityClient:
    """Implement the connectivity with the Terality API.

    This is a stateful class managing authentication and session management. Currently, a single client
    may only have one session active at any given time.

    This client is not thread-safe.

    The client can be in an "unconfigured state" (no HttpTransport associated to it). In this state,
    it will raise an exception on each request.
    The first time such an unconfigured client is called, it will log a detailed error message.

    Args:
        http_transport: the underlying HttpTransport used to perform requests. If None, then this
            client will raise a user-friendly exception on each request, explaining how to configure Terality.
        auto_session: if True, when no session is currently opened, the TeralityClient will start a new session
            before performing a request.
    """

    # Note: the session management could be split to a separate class, but for now this class
    # is small enough to keep it inline.

    def __init__(self, http_transport: Optional[HttpTransport], auto_session: bool = False) -> None:
        self._http_transport = http_transport
        self._has_printed_error_message = False
        self._session: Optional[SessionInfo] = None
        self._auto_session = auto_session
        self._credentials_fetcher = _AwsCredentialsFetcher(self)
        self._error_reporter: Optional[ErrorReporter] = None
        if self._http_transport is not None:
            self._error_reporter = ErrorReporter(self._http_transport)

    def send_request(self, route: str, payload: Optional[Any], method="POST") -> Any:
        """Send a request to the Terality API, injecting credentials and a session ID when relevant.

        The arguments are the same as the `request` method on the underlying HttpTransport.
        """
        self._raise_if_unconfigured()
        assert self._http_transport is not None

        if self._auto_session and self._session is None:
            self.start_session()

        session_id = self._session.id if self._session is not None else None

        try:
            return self._http_transport.request(
                route, payload, session_id=session_id, method=method
            )
        except Exception as e:
            if self._error_reporter is not None:
                self._error_reporter.report_exception_noexcept(e)
            raise

    def poll_for_answer(self, route: str, payload: Any, method="POST"):
        """Same as `send_request`, except that it polls the API on a PendingComputationResponse.

        On long running jobs, the API may answer with a PendingComputationResponse. This methods keeps
        polling the API until the answer is a real result.
        """
        response = self.send_request(route, payload, method=method)
        while isinstance(response, PendingComputationResponse):
            function_id = response.pending_computation_id
            response = self.send_request("follow_up", {"function_id": function_id})
        return response

    def start_session(self):
        """Start a session.

        Before processing data, the client must start a session. The data created during the session is bound
        to the session: when the session is closed, the data is garbage collected. Most API calls that deal
        with data processing require an open session.

        Raise:
            SessionStateError: if a session is already in progress
        """
        if self._session is not None:
            raise SessionStateError("A session is already opened on this client.")

        self._raise_if_unconfigured()
        assert self._http_transport is not None

        session = self._http_transport.request("create_session", payload=None)
        if not isinstance(session, CreateSessionResponse):
            raise RuntimeError(
                f"Unexpected server response type, got {type(session)}, expected SessionInfo."
            )

        self._session = SessionInfo(
            id=session.id,
            upload_config=session.upload_config,
            download_config=session.download_config,
        )

    def get_download_config(self):
        self._raise_if_unconfigured()
        if self._session is None:
            raise SessionStateError(
                "Can not get a download config on a client that does not have an active session."
            )
        return self._session.download_config

    def get_upload_config(self):
        self._raise_if_unconfigured()
        if self._session is None:
            raise SessionStateError(
                "Can not get an upload config on a client that does not have an active session."
            )
        return self._session.upload_config

    def get_aws_credentials(self) -> AwsCredentials:
        """Return AWS credentials that can be used to upload files to the AWS API.

        Don't cache these credentials anywhere as they have a short lifetime. This method already perfoms
        the necessary caching.
        """
        return self._credentials_fetcher.get_credentials()

    def close_session(self):
        """Close the current session, if any.

        If no session is in progress, this method is a no-op.
        """
        self._raise_if_unconfigured()
        if self._session is None:
            return

        assert self._http_transport is not None

        self._http_transport.request("delete_session", None, session_id=self._session.id)
        self._session = None

    def _raise_if_unconfigured(self):
        if self._http_transport is not None:
            return
        if not self._has_printed_error_message:
            self._has_printed_error_message = True
            # TODO: better error message
            logger.warning("Could not read the Terality configuration files.")
        raise UnconfiguredTeralityClientError(
            "The Terality client is not configured (more info at https://docs.terality.com/)"
        )


class UnreadableTeralityConfigError(TeralityError):
    pass


def client_from_config() -> TeralityClient:
    """Try to read the Terality configuration files, and return a configured TeralityClient from them.

    Raise:
        UnreadableTeralityConfigError: if the Terality configuration files are not accessible.

    See:
        unconfigured_client
    """
    try:
        config = TeralityConfig.load()
        credentials = TeralityCredentials.load()
        auth_credentials = AuthCredentials(
            user=credentials.user_id, password=credentials.user_password
        )
        http_transport = HttpTransport(
            base_url=config.full_url(),
            request_timeout=config.timeout,
            auth_credentials=auth_credentials,
        )
        return TeralityClient(http_transport, auto_session=True)
    except Exception as e:
        raise UnreadableTeralityConfigError("could not read Terality configuration files") from e


def unconfigured_client() -> TeralityClient:
    """
    Return an unconfigured TeralityClient. This client will raise an exception on each request.

    Useful when the caller tried to call `client_from_config`, but no config was available. This unconfigured
    client can be used to signal to the user that it needs to configure Terality.
    """
    return TeralityClient(http_transport=None, auto_session=False)


def anonymous_client(api_url: Optional[str] = None) -> TeralityClient:
    """Return an anonymous Terality client (e.g a client that does not authenticate its requests).

    Args:
        api_url: if not provided, try to get the base URL from a configuration file. If the configuration
            file is missing or unreadable, then silently use the default settings embedded in this package.
            This function can thus be safely called even on systems where Terality is not configured yet.
    """
    timeout = (1, 1)
    if api_url is None:
        config = TeralityConfig.load(fallback_to_defaults=True)
        api_url = config.full_url()
        timeout = config.timeout

    http_transport = HttpTransport(base_url=api_url, request_timeout=timeout, auth_credentials=None)
    return TeralityClient(http_transport, auto_session=False)


class _AwsCredentialsFetcher:
    """Small utility to lazily fetch temporary AWS credentials from the Terality API.

    `get_credentials` will fetch credentials on the first call, and cache the result.

    Those credentials are used to upload files to Terality-owned S3 buckets.
    """

    def __init__(self, client: TeralityClient) -> None:
        self._credentials: Optional[AwsCredentials] = None
        self._credentials_fetched_at = time.monotonic()
        self._client = client

    def get_credentials(self) -> AwsCredentials:
        if self._credentials is None or time.monotonic() > self._credentials_fetched_at + 30 * 60:
            self._fetch_credentials()
        assert self._credentials is not None
        return self._credentials

    def _fetch_credentials(self) -> None:
        res: DataTransferResponse = self._client.send_request("transfers", {})
        self._credentials = res.temporary_upload_aws_credentials
        self._credentials_fetched_at = time.monotonic()



from dataclasses import dataclass
import os
from typing import Any, Dict, Optional, Tuple
import sys
import platform

import numpy as np
import pandas as pd
from pydantic import BaseModel
import requests
from requests.adapters import Retry
from requests.models import Response
from requests.sessions import HTTPAdapter

from terality_serde import dumps, loads, TeralitySerializationError
from common_client_scheduler import headers, ErrorResponse

from terality.version import __version__
from terality.exceptions import TeralityError, TeralityClientError

_PLATFORM_INFO = platform.platform()


@dataclass
class AuthCredentials:
    user: str
    password: str


class _ProcessInfo(BaseModel):
    """Info about the Python process using Terality."""

    python_version_major: str = str(sys.version_info.major)
    python_version_minor: str = str(sys.version_info.minor)
    python_version_micro: str = str(sys.version_info.micro)
    numpy_version: str = np.__version__
    pandas_version: str = pd.__version__
    terality_version: Optional[str] = __version__
    platform: str = _PLATFORM_INFO

    def to_headers(self) -> Dict[str, str]:
        return {f"{headers.TERALITY_CLIENT_INFO_PREFIX}{k}": v for k, v in self.dict().items()}


class HttpTransport:
    """Responsible for serializing and sending requests to the Terality API, as well as deserializing the response.

    This class handles:
    * serialization and deserialization
    * raising exceptions on non-success API responses
    * retries at the HTTP level

    This class largely ignores the HTTP response code, and instead always tries to deserialize the response body.
    If this fails, it raises a generic exception.

    Args:
        base_url: URL of the Terality API (such as "https://api.terality2.com")
        verify_ssl_certificate (deprecated): whether to valide the TLS certificate chain on HTTPS connections
        request_timeout: tuple of (connect timeout, read timeout) for HTTP requests, in seconds
        auth_credentials: credentials to add to HTTP requests
    """

    def __init__(
        self,
        base_url: str,
        *,
        verify_ssl_certificate: bool = True,
        request_timeout: Tuple[int, int] = (3, 30),  # connect, read
        auth_credentials: Optional[AuthCredentials],
    ):
        self._http_session = requests.Session()
        self._base_url = base_url
        self._verify_ssl_certificate = verify_ssl_certificate
        self._request_timeout = request_timeout
        # Strip trailing slash
        if self._base_url.endswith("/"):
            self._base_url = self._base_url[:-1]

        self.set_auth_credentials(auth_credentials)
        self._process_info = _ProcessInfo()
        self._mount_http_adapter()

    def _mount_http_adapter(self):
        # We are encountering errors with the AWS API Gateway returning 404 or 403 errors
        # (ref: https://console.aws.amazon.com/support/home#/case/?displayId=8464649331&language=en)
        # While we wait for AWS support to solve this issue, also retry those calls:
        status_forcelist = ([429, 403, 404, 500, 502, 503, 504],)
        if os.environ.get("PYTEST_CURRENT_TEST"):
            # In tests, retry on less error codes to speed up tests
            status_forcelist = [403, 404]
        retry_strategy = Retry(
            total=3,
            # We are encountering errors with the AWS API Gateway returning 404 or 403 errors
            # (ref: https://console.aws.amazon.com/support/home#/case/?displayId=8464649331&language=en)
            # While we wait for AWS support to solve this issue, also retry those calls:
            status_forcelist=status_forcelist,
            # We do retry on POST methods (retry pandas computations), even if there are not idempotent.
            method_whitelist=["HEAD", "GET", "PUT", "DELETE", "OPTIONS", "TRACE", "POST"],
            backoff_factor=2,
            redirect=40,
            raise_on_redirect=True,
        )
        adapter = HTTPAdapter(max_retries=retry_strategy)
        self._http_session.mount("https://", adapter)
        self._http_session.mount("http://", adapter)

    def request(
        self,
        route: str,
        payload: Optional[Any],
        method: str = "POST",
        session_id: Optional[str] = None,
    ) -> Any:
        """Perform an API request.

        Args:
            route: URL path (such as "/compute")
            payload: object to send in the request body. Will be serialized using the `terality_serde` package.
            method: HTTP method (POST, GET, ...)
            session_id: if provided, add this session ID to the request.

        Return:
            the deserialized server response

        Raise:
            TeralityError: the server response could not be deserialized correcly
            Exception: when the server response body contains a serialized exception, it is propagated
        """
        serialized_payload = None
        if payload is not None:
            try:
                serialized_payload = dumps(payload)
            except TeralitySerializationError as e:
                if e.unserializable_type is not None:
                    raise TeralityClientError(
                        f"Serialization error: Terality does not support calling a function with a parameter of type '{e.unserializable_type}' (or a collection or object containing such a type)"
                    ) from e
                raise TeralityClientError(
                    "Serialization error: couldn't serialize the function call (unsupported parameter type?)"
                ) from e

        # Normalize route
        if not route.startswith("/"):
            route = "/" + route

        request_headers = self._make_request_headers(session_id)

        r = self._http_session.request(
            method=method,
            url=self._base_url + route,
            verify=self._verify_ssl_certificate,
            timeout=self._request_timeout,
            headers=request_headers,
            json={"session_id": session_id, "payload": serialized_payload},
        )
        return HttpTransport._deserialize_response(r)

    @staticmethod
    def _deserialize_response(r: Response) -> Any:
        # No matter the status code, we first try to parse the response body.
        # If this succeeds, we return (or raise) the deserialized object.
        try:
            parsed = loads(r.text)
        except Exception as e:  # pylint: disable=broad-except
            # If we can't parse the result, we raise a generic exception (no matter the HTTP status code).
            additional_info = HttpTransport._basic_info_from_response(r)
            raise TeralityError(r.text + " " + additional_info) from e

        # The response may also be an error. We don't use the HTTP status code here, and trust the server
        # representation instead.
        # The server representation does not include the request ID though, so we add it here.
        if isinstance(parsed, ErrorResponse):
            additional_info = HttpTransport._basic_info_from_response(r)
            raise TeralityError(f"{parsed.message} {additional_info}")

        # Otherwise, the response is not an error, just return it.
        return parsed

    def set_auth_credentials(self, auth_credentials: Optional[AuthCredentials]) -> None:
        """Set credentials for HTTP requests, or remove credentials (if None if provided)."""
        if auth_credentials is None:
            self._http_session.auth = auth_credentials
            return
        self._http_session.auth = (auth_credentials.user, auth_credentials.password)

    def _make_request_headers(self, session_id: Optional[str]) -> Dict[str, str]:
        request_headers = self._process_info.to_headers()
        if session_id:
            request_headers[headers.TERALITY_SESSION_ID] = session_id
        return request_headers

    @staticmethod
    def _extract_request_id(r: Response) -> Optional[str]:
        # The request may have failed before reaching a Terality application.
        # In that case, try to look for a request ID in headers added by AWS
        # infrastructure.
        headers_to_search = [headers.TERALITY_REQUEST_ID, "X-Amz-Apigw-Id", "Apigw-Requestid"]
        request_id = None

        for header in headers_to_search:
            request_id = r.headers.get(header)
            if request_id is not None:
                break

        return request_id

    @staticmethod
    def _basic_info_from_response(r: Response) -> str:
        request_id = HttpTransport._extract_request_id(r)
        if request_id is not None:
            info = f"(HTTP status {r.status_code}, request ID: {request_id})"
        else:
            info = f"(HTTP status {r.status_code})"
        return info



from dataclasses import dataclass
from pathlib import Path
from typing import Optional, Union

from terality_serde import SerdeMixin


@dataclass
class TransferConfigLocal(SerdeMixin):
    path: Path


@dataclass
class SessionInfoLocal(SerdeMixin):
    id: str  # pylint: disable=invalid-name
    upload_config: TransferConfigLocal
    download_config: TransferConfigLocal


@dataclass
class TransferConfig(SerdeMixin):
    default_aws_region: str
    bucket: str
    key_prefix: str

    def bucket_region(self, aws_region: Optional[str] = None) -> str:
        """Return the name of the upload bucket in the given region."""
        return f"{self.bucket}-{self.default_aws_region if aws_region is None else aws_region}"


@dataclass
class SessionInfo(SerdeMixin):
    id: str  # pylint: disable=invalid-name
    upload_config: Union[TransferConfig, TransferConfigLocal]
    download_config: Union[TransferConfig, TransferConfigLocal]



from functools import partial

from common_client_scheduler.requests_responses import ComputationResponse

from terality.exceptions import TeralityError
from terality._terality.encoding.decoder import decode
from terality_serde.recursive_apply import apply_async_func_on_object_recursively
from terality._terality.globals import global_client


def replay(bucket: str, key: str):
    """Helper to replay an API call stored in a S3 dump file.

    This function is internal to Terality, and is exposed here to help Terality engineers investigate CI
    and production failures.
    """
    response = global_client().poll_for_answer("replay", (bucket, key))

    if not isinstance(response, ComputationResponse):
        raise TeralityError(
            f"Received unexpected response type (expected ComputationResponse): {response}"
        )

    result = response.result
    result = apply_async_func_on_object_recursively(
        result, partial(decode, global_client().get_aws_credentials())
    )
    return result



from __future__ import annotations
from dataclasses import dataclass
from functools import partial
from typing import Callable
import pickle
import sys

import cloudpickle
import pandas as pd
from pandas.core.groupby import SeriesGroupBy, DataFrameGroupBy
from terality_serde.struct_types import STRUCT_TYPE_TO_PANDAS_CLASS

from . import SerdeMixin


_CLASS_NAME_TO_PANDAS_CLASS = {  # TODO replace "Index" by te.Index.__name__ etc
    "Index": pd.Index,
    "Int64Index": pd.Int64Index,
    "Float64Index": pd.Float64Index,
    "DatetimeIndex": pd.DatetimeIndex,
    "MultiIndex": pd.MultiIndex,
    "Series": pd.Series,
    "DataFrame": pd.DataFrame,
    "SeriesGroupBy": SeriesGroupBy,
    "DataFrameGroupBy": DataFrameGroupBy,
}


def _convert_terality_method_to_pandas(obj: Callable):
    """
    Need this conversion to apply the pandas function on the worker instead of the terality one.
    Exemple: when calling series.apply(te.isnull) we serde pd.isnull.

    However, this function has no effect regarding UDF using terality functions.
    For this case, the trick is to set sys.modules["terality"] = pd directly
    in the worker.
    """

    if isinstance(obj, partial):
        struct_type, accessor, method_name = obj.args
        if accessor is not None:
            raise ValueError(f"Function {obj} can not be serialized.")
        class_type = STRUCT_TYPE_TO_PANDAS_CLASS[struct_type]
    else:
        class_name = obj.__qualname__.split(".")[0]
        class_type = _CLASS_NAME_TO_PANDAS_CLASS[class_name]
        method_name = obj.__name__

    method_to_call = getattr(class_type, method_name)
    return method_to_call


@dataclass
class CallableWrapper(SerdeMixin):
    """Represent a user-supplied callable. THIS OBJECT CAN BE UNSAFE, READ THE DOCS!"""

    pickled_payload: bytes

    @classmethod
    def from_object(cls, obj: Callable) -> CallableWrapper:
        """Serialize the provided callable to a CallableWrapper.

        Internally, this uses the `cloudpickle` package and has the same limitations.

        Deserializing (even without running) such a serialized object can run arbitrary code.
        Only deserialize this in a safe sandboxed environment.
        """
        if isinstance(obj, type):
            # types are callable, but we don't support serializing them in this class
            raise TypeError("'obj' must not be a type")

        if isinstance(obj, partial):
            func = obj.func
        else:
            func = obj

        module_obj = None

        # numpy module is already impored in the worker so no need to pickle by value.
        # also, pickling it by value is buggy cf TER-563.
        if (
            hasattr(func, "__module__")
            and func.__module__ is not None
            and func.__module__ != "numpy"
        ):
            if func.__module__.split(".")[0] == "terality":
                obj = _convert_terality_method_to_pandas(obj)
                # pandas module already imported in the worker too.

            elif func.__module__ not in sys.modules:
                raise ValueError(
                    f"The provided callable module is '{func.__module__}', but it does not seem to be imported (not present in sys.modules). "
                    "Import this module in this Python session to be able to serialize this callable."
                )
            else:
                module_obj = sys.modules[func.__module__]
                cloudpickle.register_pickle_by_value(module_obj)  # type: ignore

        try:
            return cls(pickled_payload=cloudpickle.dumps(obj))  # type: ignore
        finally:
            if module_obj is not None:
                cloudpickle.unregister_pickle_by_value(module_obj)  # type: ignore

    def to_callable(self) -> Callable:
        """Return the deserialized callable. This function can run arbitrary user-supplied code and must only be run in a secure sandbox."""
        try:
            return pickle.loads(self.pickled_payload)
        except ModuleNotFoundError as e:
            raise ValueError(
                f"Can not deserialize this function, as it depends on a module not available in this execution environment: {str(e)}"
            ) from e



import os
from typing import Type, TypeVar

from abc import ABC
from pathlib import Path
from typing import ClassVar, Tuple

from pydantic import BaseModel, BaseSettings, Field  # pylint: disable=no-name-in-module


class _Paths(BaseSettings):
    terality_home: Path = Field(Path.home() / ".terality", env="TERALITY_HOME")


T = TypeVar("T", bound="BaseConfig")


class BaseConfig(BaseModel, ABC):
    """Base abstract class for configuration objects that can be stored in configuration files."""

    _rel_path: ClassVar[str]  # path to the configuration file relative to `$TERALITY_HOME`
    _permissions: ClassVar[int] = 0o644  # rw-r--r--

    @classmethod
    def file_path(cls) -> Path:
        return _Paths().terality_home / cls._rel_path

    @classmethod
    def load(cls: Type[T], fallback_to_defaults: bool = False) -> T:
        """Load the configuration object from the underlying file.

        Args:
            fallback_to_defaults: if True, return a default configuration file when the file does
                not already exists, or if any error occurs when reading it

        Return:
            a configuration object

        Raise:
            Exception: if the file can not be found (when fallback_to_defaults is False)
        """
        file_path = cls.file_path()
        try:
            return cls.parse_file(file_path)
        except Exception:  # pylint: disable=broad-except
            # TODO: use a less generic exception
            if not fallback_to_defaults:
                raise
            return cls()

    def save(self) -> None:
        file_path = self.file_path()
        file_path.parent.mkdir(parents=True, exist_ok=True)
        with open(
            os.open(file_path, os.O_WRONLY | os.O_CREAT | os.O_TRUNC, mode=self._permissions),
            "w",
        ) as f:
            f.write(self.json(indent=4))


class TeralityCredentials(BaseConfig):
    """User authentication information."""

    _rel_path: ClassVar[str] = "credentials.json"
    _permissions: ClassVar[int] = 0o600  # rw-------
    user_id: str  # Client-facing user ID (probably an email). Note that the "server user ID" is different.
    user_password: str  # API key or similar


_PRODUCTION_URL = "api.terality2.com/v1"


class TeralityConfig(BaseConfig):
    """Generic configuration related to the Terality API.

    The default values are suitable for the production SaaS environment.
    """

    _rel_path: ClassVar[str] = "config.json"
    url: str = _PRODUCTION_URL
    use_https: bool = True
    requests_ssl_verification: bool = (
        True  # note: this parameter is mostly ignored right now, and will probably be deprecated
    )
    timeout: Tuple[int, int] = (3, 35)  # The server has a max 30s timeout on responses.
    is_demo: bool = False

    # Legacy - the URL parameter should contain everything, but we don't want to invalidate all
    # existing configuration files, nor (for now) deal with multiple configuration versions.
    def full_url(self):
        prefix = "https" if self.use_https else "http"
        return f"{prefix}://{self.url}"

    @property
    def ignore_session(self) -> bool:
        return self.is_demo

    @property
    def skip_transfers(self) -> bool:
        return self.is_demo

    @property
    def is_production(self) -> bool:
        return self.url == _PRODUCTION_URL



"""Utilities to report client errors to Sentry (sentry.io).

See the Sentry docs for details.
"""
import os
from typing import Optional

import sentry_sdk
from sentry_sdk.integrations.excepthook import ExcepthookIntegration
import numpy as np
import pandas as pd

from .constants import ENV_PROD, TERALITY_DISABLE_TELEMETRY_ENV_VAR


def _drop_non_prod_events(event, hint):  # pylint: disable=unused-argument
    # Allows for explicitly disabling Sentry.
    if os.environ.get(TERALITY_DISABLE_TELEMETRY_ENV_VAR, "0") == "1":
        return None
    # Only send events in production. Don't clutter Sentry with other events.
    if event["environment"] != ENV_PROD:
        return None

    # Never report KeyboardInterrupt.
    if "exc_info" in hint:
        _exc_type, exc_value, _traceback = hint["exc_info"]
        if isinstance(exc_value, KeyboardInterrupt):
            return None

    # Only report exceptions that have somehow a source in the Terality module.
    # (we are not interested in exceptions from the user code outside Terality, but Sentry applies
    # to the whole Python session, e.g the whole Jupyter Lab).
    # We can't inspect the top-level "module" key in the Sentry event, because this module may not
    # be "terality", but one called by terality (e.g "socket" for network issues).
    # Instead, inspect the stack trace, and check that at some point the Terality module appears.
    try:
        values = event["exception"]["values"]
        for value in values:
            frames = value["stacktrace"]["frames"]
            for frame in frames:
                if frame["module"].split(".")[0] == "terality":
                    return event
    except KeyError:
        # Drop events with an unexpected schema - they probably don't interest us.
        return None
    return None


def set_up_sentry(environment: str, release: Optional[str], user_id: Optional[str] = None) -> None:
    """Configure the Sentry integration (for client-side error reporting).

    If the user ID is not known when calling this method, it can be set later with `set_user_for_sentry`.
    """
    if release is None:
        release = "unknown"
    # This Sentry DSN is not security sensitive.
    sentry_sdk.init(
        "https://478ff7ea710140e895b0c8734b9fb802@o923608.ingest.sentry.io/5871105",
        traces_sample_rate=1.0,
        release=release,
        environment=environment,
        before_send=_drop_non_prod_events,
        # By default, Sentry would not send events coming from a REPL.
        # However, a Jupyter notebook is a REPL, and we do want events from the Jupyter notebook.
        # The _drop_non_prod_events function should take care of removing simple typos and other
        # uninteresting events from Sentry events.
        integrations=[ExcepthookIntegration(always_run=True)],
    )
    if user_id is not None:
        set_user_for_sentry(user_id)
    set_sentry_tags()


def set_user_for_sentry(user_id: str):
    """Associate the current Terality user ID to Sentry events."""
    sentry_sdk.set_user({"id": user_id, "username": user_id})


def set_sentry_tags():
    # Sentry does send these versions as part of the event context, but settings them as tags
    # allows for searching issues by these versions (Sentry context is not searchable).
    sentry_sdk.set_tag("versions.numpy", np.__version__)
    sentry_sdk.set_tag("versions.pandas", pd.__version__)



from functools import partial
from typing import Optional, Any, Union

from terality_serde import apply_func_on_object_recursively, StructType, IndexType
from common_client_scheduler import (
    ComputationResponse,
    PandasFunctionRequest,
)

from terality.exceptions import TeralityError
from .. import global_client, encode, decode


def call_pandas_function(
    function_type: Union[StructType, IndexType],
    function_prefix: Optional[str],
    function_name: str,
    *args,
    **kwargs,
) -> Any:

    args = [] if args is None else args
    args_encoded = apply_func_on_object_recursively(
        args, partial(encode, global_client().get_aws_credentials(), function_name)
    )
    kwargs = {} if kwargs is None else kwargs
    kwargs_encoded = apply_func_on_object_recursively(
        kwargs, partial(encode, global_client().get_aws_credentials(), function_name)
    )
    job = PandasFunctionRequest(
        function_type=function_type,
        function_accessor=function_prefix,
        function_name=function_name,
        args=args_encoded,
        kwargs=kwargs_encoded,
    )

    response = global_client().poll_for_answer("compute", job)

    if not isinstance(response, ComputationResponse):
        raise TeralityError(f"Received unexpected response {response}")

    result = response.result
    result = apply_func_on_object_recursively(
        result, partial(decode, global_client().get_aws_credentials())
    )

    # Handle in-place modification of data structures.
    # NOTE: Because `inplace` is only available on methods, the first argument is guaranteed to be positional
    # (`self`).
    if response.inplace:
        from terality._terality.terality_structures.structure import (
            Struct,
        )  # break cyclic import

        if not isinstance(args[0], Struct):
            raise TeralityError("Received in-place response but the target is not a data structure")
        target = args[0]
        if not isinstance(result, Struct):
            raise TeralityError("Received in-place response but the result is not a data structure")
        # noinspection PyProtectedMember
        target._mutate(result)
        result = None
    return result



from pathlib import Path
from typing import List

from common_client_scheduler.requests_responses import AwsCredentials

from .common import DataTransfer
from .. import global_client
from ...exceptions import TeralityClientError


def upload_local_files(path: str, transfer_id: str, aws_credentials: AwsCredentials) -> None:
    """
    Copy files from a local directory to a Terality-owned S3 bucket.

    Args:
        path: path to a single file or a directory. If a directory, all files in the directory will be uploaded.
    """
    try:
        paths: List[str] = (
            [path]
            if Path(path).is_file()
            else [str(path_) for path_ in sorted(Path(path).iterdir())]
        )
        for file_num, _ in enumerate(paths):
            DataTransfer.upload_local_file(
                global_client().get_upload_config(),
                aws_credentials,
                paths[file_num],
                f"{transfer_id}/{file_num}.data",
            )
    except FileNotFoundError as e:
        raise TeralityClientError(
            f"File '{path}' could not be found in your local directory, please verify the path. If your file is stored on the cloud, make sure your path starts with 's3://', 'abfs://', or 'az://'.",
        ) from e



import os
from dataclasses import dataclass
from io import BytesIO
from pathlib import Path
from typing import Any, Iterator, Optional, Tuple
from uuid import uuid4

import boto3
from botocore import UNSIGNED
from botocore.client import BaseClient
from botocore.config import Config

from common_client_scheduler import AwsCredentials
from terality.exceptions import TeralityClientError

OBJECT_CHUNK_SIZE_BYTES = 200 * 1024 * 1024  # objects to copy are copied chunk by chunk


class S3:
    # Authentication
    # ~~~~~~~~~~~~~~
    #
    # We use the S3 client in two configurations:
    # * local to S3 copies, or S3 to local copies (where S3 is a Terality owned bucket)
    # * S3 to S3 copies (between a user and a Terality owned bucket)
    #
    # While the user may have her own AWS credentials in her environment, we can't use them to write to
    # Terality buckets. Indeed, regardless of bucket policies, the user won't be able to perform a PutObject
    # (or CreateMultipartUpload) on a Terality bucket without a policy giving him "s3:PutObject" on said
    # bucket. Same goes for reads.
    #
    # Instead, whenever the client needs to perform a write or read operation on a Terality bucket,
    # it retrieves short-lived, temporary AWS credentials from the Terality API suitable for the operation.
    #
    # This is simpler than having the server generate pre-signed URLs, and provides the same level of
    # security.
    #
    # However, Terality-issued credentials can't be used to read or write S3 files to user buckets.
    #
    # We force the signature_version to s3v4 (signature version 4) to avoid edge cases where the user
    # would have set another default signature version.
    _client: Any = None
    _client_with_credentials: Any = None
    _client_anonymous: Any = None
    _credentials: Optional[AwsCredentials] = None

    @classmethod
    def client(cls):
        """An S3 client using user credentials (using the default boto3 loading mecanism).

        Operations with this client will fail if the user has no AWS credentials available in their
        environment.
        """
        if cls._client is None:
            cls._client = boto3.session.Session().client(
                "s3", config=Config(max_pool_connections=200, signature_version="s3v4")
            )
        return cls._client

    @classmethod
    def client_from_credentials(cls, credentials: AwsCredentials):
        """An authenticated S3 client, with temporary credentials provided by Terality."""
        if cls._credentials != credentials:
            cls._client_with_credentials = boto3.session.Session().client(
                "s3",
                config=Config(max_pool_connections=200, signature_version="s3v4"),
                aws_access_key_id=credentials.access_key_id,
                aws_secret_access_key=credentials.secret_access_key,
                aws_session_token=credentials.session_token,
            )
            cls._credentials = credentials
        return cls._client_with_credentials

    @classmethod
    def client_anonymous(cls):
        """An anonymous S3 client, since the user IAM role may have permission blocked to
        s3 buckets not in his account"""
        if cls._client_anonymous is None:
            cls._client_anonymous = boto3.session.Session().client(
                "s3", config=Config(signature_version=UNSIGNED)
            )
        return cls._client_anonymous


_ACL = {"ACL": "bucket-owner-full-control"}


class DataTransfer:
    """
    Various functions to upload and download files to/from S3.

    These functions are defined as class methods to make writing mocks easier during testing.
    """

    @staticmethod
    def upload_bytes(upload_config, aws_credentials: AwsCredentials, data: BytesIO) -> str:
        data.seek(0)
        transfer_id = str(uuid4())
        key = f"{upload_config.key_prefix}{transfer_id}/0.data"
        S3.client_from_credentials(aws_credentials).upload_fileobj(
            Fileobj=data, Bucket=upload_config.bucket_region(), Key=key, ExtraArgs=_ACL
        )
        return transfer_id

    @staticmethod
    def upload_local_file(
        upload_config,
        aws_credentials: AwsCredentials,
        local_file: str,
        file_suffix: str,
    ) -> None:
        key = f"{upload_config.key_prefix}{file_suffix}"
        S3.client_from_credentials(aws_credentials).upload_file(
            Bucket=upload_config.bucket_region(),
            Key=key,
            Filename=local_file,
            ExtraArgs=_ACL,
        )

    @staticmethod
    def download_to_bytes(
        download_config, aws_credentials: AwsCredentials, transfer_id: str
    ) -> BytesIO:
        buf = BytesIO()
        key = f"{download_config.key_prefix}{transfer_id}/0.parquet"
        S3.client_from_credentials(aws_credentials).download_fileobj(
            Bucket=download_config.bucket_region(), Key=key, Fileobj=buf
        )
        return buf

    @staticmethod
    def download_to_local_files(
        download_config,
        aws_credentials: AwsCredentials,
        transfer_id: str,
        path: str,
        is_folder: bool,
        with_leading_zeros: bool,
    ) -> None:
        """
        Either download one or several files if to_xxx_folder was called.
        path basename must have a '*' character here, this check is performed earlier.
        """

        bucket = download_config.bucket_region()
        key_prefix = f"{download_config.key_prefix}{transfer_id}/"
        client = S3.client_from_credentials(aws_credentials)
        keys = [o[0] for o in list_all_objects_under_prefix(client, bucket, key_prefix)]
        if is_folder:
            dirname = os.path.dirname(path)
            filename = os.path.basename(path)
            Path(dirname).mkdir(parents=True, exist_ok=True)
            for num, key in enumerate(keys):
                file_num = f"{num:0{len(str(len(keys)))}d}" if with_leading_zeros else str(num + 1)
                client.download_file(
                    Bucket=bucket,
                    Key=key,
                    Filename=f"{dirname}/{filename.replace('*', file_num)}",
                )
        else:
            client.download_file(Bucket=bucket, Key=keys[0], Filename=path)


def list_all_objects_under_prefix(
    client: BaseClient, s3_bucket: str, s3_key_prefix: str
) -> Iterator[Tuple[str, int]]:
    """
    List all S3 keys in a bucket starting with a given prefix.

    Yield:
        tuples of (S3 object key, object size in bytes)
    """
    paginator = client.get_paginator("list_objects_v2")
    page_iterator = paginator.paginate(Bucket=s3_bucket, Prefix=s3_key_prefix)
    try:
        for page in page_iterator:
            objects = page["Contents"]
            for obj in objects:
                yield (obj["Key"], obj["Size"])
    except KeyError as e:
        raise TeralityClientError(
            "Encountered an empty prefix on S3, please check that the path you have entered is valid"
        ) from e


@dataclass
class ExportedS3File:
    """A S3 object stored in a Terality export bucket."""

    bucket: str
    key: str
    size_bytes: int



import numpy as np
import pandas as pd
from pandas.core.groupby import SeriesGroupBy, DataFrameGroupBy
from terality_serde import SerializableEnum


# This class needs to be common to client / scheduler / worker.
class IndexType(str, SerializableEnum):
    """
    Technical types names of supported index sub classes.
    """

    INDEX = "index"
    MULTI_INDEX = "multi_index"
    DATETIME_INDEX = "datetime_index"
    INT64_INDEX = "int64_index"
    FLOAT64_INDEX = "float64_index"


class StructType(str, SerializableEnum):
    """
    Technical types names of pandas supported structures, excluding index.
    """

    DATAFRAME = "dataframe"
    SERIES = "series"
    DATAFRAME_GROUPBY = "dataframe_groupby"
    SERIES_GROUPBY = "series_groupby"
    NDARRAY = "ndarray"
    TOP_LEVEL = "top_level"


STRUCT_TYPE_TO_PANDAS_CLASS = {
    IndexType.INDEX: pd.Index,
    IndexType.INT64_INDEX: pd.Int64Index,
    IndexType.FLOAT64_INDEX: pd.Float64Index,
    IndexType.DATETIME_INDEX: pd.DatetimeIndex,
    IndexType.MULTI_INDEX: pd.MultiIndex,
    StructType.NDARRAY: np.ndarray,
    StructType.SERIES: pd.Series,
    StructType.DATAFRAME: pd.DataFrame,
    StructType.SERIES_GROUPBY: SeriesGroupBy,
    StructType.DATAFRAME_GROUPBY: DataFrameGroupBy,
    StructType.TOP_LEVEL: pd,
}



from dataclasses import dataclass
import json
from typing import List, Optional, Union

import numpy as np
import pandas as pd
import pyarrow

from terality_serde import loads, dumps, SerdeMixin, SerializableEnum


def instantiate_pandas_arrow_extension_types():
    """
    Pyarrow has a weird bug where for the following types, if you read a parquet using these types directly
    it will fail to load to the right pandas extension dtypes. However, if you already used these types
    manually, they will be "available" and the parquet loading will work fine
    """
    data_to_instantiate = [
        pd.arrays.PeriodArray([1], freq="M"),
        [pd.Interval(pd.Timestamp.now(), pd.Timestamp.now(), closed="left")],
    ]
    for data in data_to_instantiate:
        pyarrow.Table.from_pandas(pd.Series(data).to_frame("foo"))


class ArrayEncoding(SerializableEnum):
    MIXED = "MIXED"
    STRUCT = "STRUCT"
    DURATION = "DURATION"


JSONABLE_TYPES = (float, int, bool, str, type(None))


@dataclass
class DFParquetEncoding(SerdeMixin):
    index: List[Optional[ArrayEncoding]]
    cols: List[Optional[ArrayEncoding]]


def has_float_nan(array: Union[np.ndarray, pd.array]) -> bool:
    series = pd.Series(array)
    return series[series.isna()].apply(lambda x: isinstance(x, float)).any()


def has_only_jsonable_types(array: np.ndarray):
    series = pd.Series(array)
    return series.apply(lambda x: isinstance(x, JSONABLE_TYPES)).all()


def decode_struct_array(array: pd.array) -> pd.array:
    """
    - Can't use apply to avoid pandas auto-casting.
    - Can't use pd.array(data) because if data has tuples
      pandas thinks we want to build a 2D array.
    """
    data_decoded = [loads(val) for val in array]
    return pd.Series(data_decoded, dtype="object").array


def decode_mixed_array(array: pd.array) -> pd.array:
    """
    - Can't use apply to avoid pandas auto-casting.
    - Can't use pd.array(data) because if data has tuples
      pandas thinks we want to build a 2D array.
    """
    data_decoded = [json.loads(val) for val in array]
    return pd.Series(data_decoded, dtype="object").array


def encode_struct_array(array: pd.array) -> pd.array:
    """
    Force dtype to avoid pandas auto-casting (ex: datetime.datetime to pd.Timestamp).
    """

    return pd.Series(array, dtype=array.dtype).apply(dumps).array


def encode_mixed_array(array: pd.array) -> pd.array:
    """
    Force dtype to avoid potential pandas auto-casting.
    """

    return pd.Series(array, dtype=array.dtype).apply(json.dumps).array


_encodings = {
    ArrayEncoding.MIXED: encode_mixed_array,
    ArrayEncoding.STRUCT: encode_struct_array,
    # pyarrow to parquet does not support duraiont[ns]
    ArrayEncoding.DURATION: lambda x: x.view("int64"),  # pandas says view and not astype
}

_decodings = {
    ArrayEncoding.MIXED: decode_mixed_array,
    ArrayEncoding.STRUCT: decode_struct_array,
    ArrayEncoding.DURATION: lambda array: pd.array(array.astype("m8[ns]")),
}


def encode_array_for_parquet(
    array: pd.api.extensions.ExtensionArray, array_encoding: Optional[ArrayEncoding], encode: bool
) -> pd.api.extensions.ExtensionArray:
    if array_encoding is None:
        return array

    # allow not raising on unknown encodings to not break the compatibility with old clients when we add
    # new encodings
    encoder = _encodings.get(array_encoding) if encode else _decodings.get(array_encoding)
    if encoder is None:
        return array

    return encoder(array)


def _endecode_json_indices_inplace(
    df: pd.DataFrame, indices_encoding: List[Optional[ArrayEncoding]], encode: bool
) -> None:
    """
    Encode/Decode index levels which are either Mixed or Struct types.
    Mutates the input DataFrame.
    """
    if isinstance(df.index, pd.RangeIndex):
        # enconding will convert RangeIndex to Int64Index
        return

    assert len(indices_encoding) == df.index.nlevels
    new_arrays = [
        encode_array_for_parquet(df.index.get_level_values(level_num).array, index_encoding, encode)
        for level_num, index_encoding in enumerate(indices_encoding)
    ]

    if isinstance(df.index, pd.MultiIndex):
        df.index = pd.MultiIndex.from_arrays(new_arrays, names=df.index.names)
    else:
        assert len(new_arrays) == 1
        df.index = pd.Index(new_arrays[0], name=df.index.name)


def _endecode_json_columns_inplace(
    df: pd.DataFrame, cols_encoding: List[Optional[ArrayEncoding]], encode: bool
) -> None:
    """
    Encode/Decode columns which are either Mixed or Struct types.
    Mutates the input DataFrame.
    """
    assert len(cols_encoding) == len(df.columns)
    for col_num, col_need_encoding in enumerate(cols_encoding):
        array = df.iloc[:, col_num].array
        df.iloc[:, col_num] = encode_array_for_parquet(array, col_need_encoding, encode)


def decode_df_inplace(df: pd.DataFrame, encoding: DFParquetEncoding) -> None:
    _endecode_json_indices_inplace(df, encoding.index, encode=False)
    _endecode_json_columns_inplace(df, encoding.cols, encode=False)


def encode_df_inplace(df: pd.DataFrame, encoding: DFParquetEncoding) -> None:
    _endecode_json_indices_inplace(df, encoding.index, encode=True)
    _endecode_json_columns_inplace(df, encoding.cols, encode=True)



"""Azure integration. Don't import this module at the top level of another module.

This module depends on the Azure Python SDK which is an extra (optional) depedency. You need to handle
import errors when you try to import this module.
"""
from datetime import datetime, timedelta
import logging
import os

from azure.storage.blob._shared.models import UserDelegationKey
from terality._terality.data_transfers.common import (
    ExportedS3File,
    OBJECT_CHUNK_SIZE_BYTES,
    S3,
    list_all_objects_under_prefix,
)
from typing import Dict, Iterator, List, Optional, Tuple, Union, cast

from azure.identity import DefaultAzureCredential
from azure.storage.blob import BlobServiceClient
from azure.storage.blob import generate_blob_sas
from azure.storage.filedatalake import FileSystemClient

from common_client_scheduler.requests_responses import (
    AwsCredentials,
    AzureBlobSource,
    AzureBlobStoragePartExportRequest,
    AzureBlobStoragePartsExport,
    ExportToCloudRequest,
    ImportFromAzureBlobStorageSource,
    ImportFromCloudRequest,
    ObjectStorageKey,
    StorageService,
)

from .. import global_client

logger = logging.getLogger("azure")
logger.setLevel(logging.ERROR)


def upload_azure_storage_files(storage_account_name: str, container: str, folder: str) -> str:
    credential, storage_account_key = _get_azure_storage_credentials()

    blob_account_url, datalake_account_url = _storage_account_urls_from_name(storage_account_name)
    blob_service_client = BlobServiceClient(account_url=blob_account_url, credential=credential)
    start_time, expiry_time = _get_signature_time_range()

    signature = _get_signature_key(
        blob_service_client, storage_account_key, start_time, expiry_time
    )
    filesystem_client = FileSystemClient(datalake_account_url, container, credential=credential)
    objects_to_copy = list(
        list_objects_to_copy(
            container=container,
            folder=folder,
            storage_account_name=storage_account_name,
            expiry_time=expiry_time,
            start_time=start_time,
            signature=signature,
            blob_service_client=blob_service_client,
            filesystem_client=filesystem_client,
        )
    )
    return _send_transfer_request(objects_to_copy)


def _get_signature_time_range() -> Tuple[datetime, datetime]:
    start_time = datetime.utcnow()
    expiry_time = datetime.utcnow() + timedelta(hours=1)
    return start_time, expiry_time


def _storage_account_urls_from_name(storage_account_name: str) -> Tuple[str, str]:
    datalake_account_url = f"https://{storage_account_name}.dfs.core.windows.net"
    blob_account_url = f"https://{storage_account_name}.blob.core.windows.net"
    return blob_account_url, datalake_account_url


def _get_azure_storage_credentials() -> Tuple[Union[str, DefaultAzureCredential], Optional[str]]:
    storage_account_key = os.environ.get("AZURE_STORAGE_ACCOUNT_KEY")
    if storage_account_key is not None:
        credential = storage_account_key
    else:
        credential = DefaultAzureCredential()
    return credential, storage_account_key


def list_objects_to_copy(
    *,
    storage_account_name: str,
    container: str,
    folder: str,
    signature: Union[Dict[str, UserDelegationKey], Dict[str, str]],
    expiry_time: datetime,
    start_time: datetime,
    blob_service_client: BlobServiceClient,
    filesystem_client: FileSystemClient,
) -> Iterator[AzureBlobSource]:
    for path in _list_blobs(filesystem_client, folder):
        shared_access_signature = generate_blob_sas(
            account_name=storage_account_name,
            container_name=container,
            blob_name=path,
            permission="r",
            expiry=expiry_time,
            start_time=start_time,
            **signature,
        )
        blob_client = blob_service_client.get_blob_client(container, path)
        blob_size = cast(int, blob_client.get_blob_properties().size)
        for (range_start_byte, range_end_byte) in _make_chunks(blob_size):
            yield AzureBlobSource(
                object_key=ObjectStorageKey(bucket=container, key=path),
                shared_access_signature=shared_access_signature,
                storage_account_name=storage_account_name,
                range_start_byte=range_start_byte,
                range_end_byte=range_end_byte,
            )


def _get_signature_key(
    blob_service_client: BlobServiceClient,
    storage_account_key: Optional[str],
    start_time: datetime,
    expiry_time: datetime,
) -> Union[Dict[str, UserDelegationKey], Dict[str, str]]:
    if storage_account_key is None:
        user_delegation_key = blob_service_client.get_user_delegation_key(
            key_start_time=start_time, key_expiry_time=expiry_time
        )
        signature = {"user_delegation_key": user_delegation_key}
    else:
        signature = {"account_key": storage_account_key}
    return signature


def _send_transfer_request(objects_to_copy: List[AzureBlobSource]) -> str:
    request = ImportFromCloudRequest(
        service=StorageService.AZURE_BLOB_STORAGE,
        source=ImportFromAzureBlobStorageSource(objects=objects_to_copy),
    )
    response = global_client().poll_for_answer("imports/cloud", request)
    return response.transfer_id


def _list_blobs(client: FileSystemClient, folder: str) -> Iterator[str]:
    for path in client.get_paths(folder):
        if path.is_directory:
            yield from _list_blobs(client, path)
        else:
            yield path.name


def _make_chunks(
    object_size: int, chunk_size: int = OBJECT_CHUNK_SIZE_BYTES
) -> Iterator[Tuple[int, int]]:
    """Split the given size in consecutive ranges of chunk_size bytes (offsets are inclusive)."""
    range_start = 0
    while range_start < object_size:
        range_end = min(range_start + chunk_size, object_size) - 1
        yield (range_start, range_end)
        range_start = range_end + 1


def copy_to_azure_datalake(
    *,
    aws_credentials: AwsCredentials,
    transfer_id: str,
    aws_region: Optional[str],
    storage_account_name: str,
    container: str,
    path_template: str,
) -> None:
    """Copy the files identified by `transfer_id` from a Terality export bucket to an Azure Datalake Gen2 filesystem..

    Args:
        aws_credentials: AWS credentials to use to read files in the Terality export bucket
        transfer_id: transfer ID used to identify the files to copy from the Terality export bucket
        aws_region: AWS region hosting the Terality export bucket
        storage_account_name: destination Azure storage account
        container: destination Azure container
        path_template: destination path, with the pattern "some/path/*.extension". The '*' will be replaced by integers
            (starting from 0) in the destination.
    """
    download_config = global_client().get_download_config()
    source_bucket = download_config.bucket_region(aws_region)
    source_prefix = f"{download_config.key_prefix}{transfer_id}/"

    auth_s3_client = S3.client_from_credentials(aws_credentials)
    files = [
        (
            path_template.replace("*", str(i)),
            ExportedS3File(bucket=source_bucket, key=s3_key, size_bytes=object_size),
        )
        for i, (s3_key, object_size) in enumerate(
            list_all_objects_under_prefix(auth_s3_client, source_bucket, source_prefix)
        )
    ]
    _copy_exported_files_to_azure_destination(files, storage_account_name, container)


def _copy_exported_files_to_azure_destination(
    files: List[Tuple[str, ExportedS3File]], storage_account_name: str, container: str
) -> None:
    credential, storage_account_key = _get_azure_storage_credentials()
    blob_account_url, _ = _storage_account_urls_from_name(storage_account_name)
    blob_service_client = BlobServiceClient(account_url=blob_account_url, credential=credential)
    start_time, expiry_time = _get_signature_time_range()
    signature = _get_signature_key(
        blob_service_client, storage_account_key, start_time, expiry_time
    )

    to_copy = list(
        _generate_blob_parts(
            files=files,
            storage_account_name=storage_account_name,
            container=container,
            expiry_time=expiry_time,
            start_time=start_time,
            signature=signature,
        )
    )
    request = ExportToCloudRequest(
        service=StorageService.AZURE_BLOB_STORAGE,
        export_request=AzureBlobStoragePartsExport(parts=to_copy),
    )
    global_client().poll_for_answer("exports/cloud", request)


def _generate_blob_parts(
    *,
    files: List[Tuple[str, ExportedS3File]],
    storage_account_name: str,
    container: str,
    expiry_time: datetime,
    start_time: datetime,
    signature: Dict,
) -> Iterator[AzureBlobStoragePartExportRequest]:
    for (destination_path, s3_file) in files:
        shared_access_signature = generate_blob_sas(
            account_name=storage_account_name,
            container_name=container,
            blob_name=destination_path,
            permission="w",
            expiry=expiry_time,
            start_time=start_time,
            **signature,
        )
        for (range_start_byte, range_end_byte) in _make_chunks(s3_file.size_bytes):
            yield AzureBlobStoragePartExportRequest(
                source_object_key=ObjectStorageKey(s3_file.bucket, s3_file.key),
                range_start_byte=range_start_byte,
                range_end_byte=range_end_byte,
                destination_object_key=ObjectStorageKey(bucket=container, key=destination_path),
                storage_account_name=storage_account_name,
                shared_access_signature=shared_access_signature,
            )



"""This module defines global instances created when importing Terality.

Importing the terality module should have as few side-effects as possible, and they should not
occur outside of the module.
"""
import atexit
import os
import sys

import requests

from terality.exceptions import TeralityError
from terality.version import __version__
from .client import (
    TeralityClient,
    client_from_config,
    UnreadableTeralityConfigError,
    unconfigured_client,
)
from .constants import ENV_PROD, ENV_TESTS, ENV_UNKNOWN
from .utils import logger, TeralityConfig
from .utils.config import TeralityCredentials
from .sentry import set_up_sentry


def _latest_version_from_pypi() -> str:
    root_package_name = __name__.split(".")[0]
    r = requests.get(f"https://pypi.org/pypi/{root_package_name}/json")
    r.raise_for_status()
    return r.json()["info"]["version"]


def _print_warning_if_not_latest_version() -> None:
    try:
        # Don't perform useless network requests during tests
        if "PYTEST_CURRENT_TEST" in os.environ:
            return
        latest = _latest_version_from_pypi()
        if latest != __version__:
            logger.warning(
                f"You are using version {__version__} of the Terality client, but version {latest} is available. "
                "Consider upgrading your version to get the latest fixes and features."
            )
    except Exception:  # pylint:disable=broad-except  # nosec
        # If any error occurs, don't write a stack trace, just swallow the exception.
        pass


def _atexit_delete_session():
    """Try to close the current global session, using a best effort policy. Swallow any Terality exception."""
    try:
        global_client().close_session()
    except TeralityError:
        pass  # nosec: B110 (try_except_pass)


def global_client() -> TeralityClient:
    """Return a preconfigured Terality client.

    If no configuration is available on the system running Terality code, then this client may not be configured.
    In that case, any attempt to call a method on the client will raise an exception.

    This global Terality client is defined in order to hide the existence of a TeralityClient class to
    the user.

    Code in the Terality module can also manually instantiate a TeralityClient with a different
    configuration when required. For instance, an anonymous client (without credentials) can be useful
    for some requests. The `terality._terality.client` moduule contains several helpers to do
    just that.
    """
    return _GLOBAL_CLIENT


def load_global_client():
    """Load the Terality configuration. If a previous configuration was loaded, discard it.

    This method should be called anytime the configuration files change while the Python interpreter is running.
    Otherwise, the new configuration won't be taken into account.
    """
    # Try to get a client from configuration files.
    global _GLOBAL_CLIENT  # pylint: disable=global-statement
    try:
        _GLOBAL_CLIENT = client_from_config()
    except UnreadableTeralityConfigError:
        # If this fails, default to a client that will warn the user on requests.
        # It's important not to log anything at this point, as the Terality module may be imported in places
        # where not having such a configuration is expected.
        _GLOBAL_CLIENT = unconfigured_client()


def _init_sentry():
    config = TeralityConfig().load(fallback_to_defaults=True)
    environment = ENV_PROD if config.is_production else ENV_UNKNOWN
    # If the pytest module has already been loaded, we assume we're running in some kind of tests
    # environment, to avoid sending events to Sentry.
    # This is not fool-proof, but there is not great heuristic here.
    # (alternative: test if sys.argv[0] == "pytest", but this may not work for pytest wrappers).
    if "pytest" in sys.modules:
        environment = ENV_TESTS
    try:
        creds = TeralityCredentials().load()
        user_id = creds.user_id
    except Exception:  # pylint: disable=broad-except
        # TODO: less broad exception type
        user_id = None

    set_up_sentry(environment=environment, release=__version__, user_id=user_id)


# Importing the Terality module has some side-effects, defined below.
# This is a consequence of sticking to the pandas API: we can't require the user to manually instantiate a
# "client" class or similar.

_init_sentry()
_print_warning_if_not_latest_version()

_GLOBAL_CLIENT = unconfigured_client()
# Try to replace the _GLOBAL_CLIENT variable with a configured client at module import time.
load_global_client()
atexit.register(_atexit_delete_session)
